# Copyright (c) Acconeer AB, 2022-2025
# All rights reserved

from __future__ import annotations

import logging
from enum import Enum, auto
from typing import Any, Callable, Mapping, Optional

import attrs
import h5py
import numpy as np

from PySide6 import QtCore
from PySide6.QtWidgets import QPushButton, QVBoxLayout

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool import a121
from acconeer.exptool.a121._h5_utils import _create_h5_string_dataset
from acconeer.exptool.a121.algo._plugins import (
    A121BackendPluginBase,
    A121ViewPluginBase,
)
from acconeer.exptool.app.new import (
    AppModel,
    AttrsConfigEditor,
    BackendLogger,
    GeneralMessage,
    GroupBox,
    Message,
    MiscErrorView,
    PgPlotPlugin,
    PidgetGroupFactoryMapping,
    PluginFamily,
    PluginGeneration,
    PluginPresetBase,
    PluginSpecBase,
    backend,
    icons,
    is_task,
    pidgets,
    visual_policies,
)
from acconeer.exptool.app.new.ui.components import CollapsibleWidget, GotoResourceTabButton
from acconeer.exptool.app.new.ui.components.a121 import (
    SensorConfigEditor,
)
from acconeer.exptool.app.new.ui.components.json_save_load_buttons import PresentationType

from . import _pidget_mapping
from ._configs import (
    get_long_range_config,
    get_low_power_config,
    get_medium_range_config,
    get_short_range_config,
)
from ._detector import (
    Detector,
    DetectorConfig,
    DetectorContext,
    DetectorMetadata,
    DetectorResult,
    _load_algo_data,
)


log = logging.getLogger(__name__)


@attrs.mutable(kw_only=True)
class SharedState:
    sensor_id: int = attrs.field(default=1)
    config: DetectorConfig = attrs.field(factory=DetectorConfig)
    context: Optional[DetectorContext] = attrs.field(default=None)


class PluginPresetId(Enum):
    SHORT_RANGE = auto()
    MEDIUM_RANGE = auto()
    LONG_RANGE = auto()
    LOW_POWER = auto()


@attrs.frozen(kw_only=True)
class SetupMessage(GeneralMessage):
    detector_config: DetectorConfig
    detector_metadata: DetectorMetadata
    estimated_frame_rate: float
    name: str = attrs.field(default="setup", init=False)
    recipient: backend.RecipientLiteral = attrs.field(default="plot_plugin", init=False)


class BackendPlugin(A121BackendPluginBase[SharedState]):
    PLUGIN_PRESETS: Mapping[int, Callable[[], DetectorConfig]] = {
        PluginPresetId.SHORT_RANGE.value: lambda: get_short_range_config(),
        PluginPresetId.MEDIUM_RANGE.value: lambda: get_medium_range_config(),
        PluginPresetId.LONG_RANGE.value: lambda: get_long_range_config(),
        PluginPresetId.LOW_POWER.value: lambda: get_low_power_config(),
    }

    def __init__(
        self, callback: Callable[[Message], None], generation: PluginGeneration, key: str
    ) -> None:
        super().__init__(callback=callback, generation=generation, key=key)

        self._recorder: Optional[a121.H5Recorder] = None
        self._detector_instance: Optional[Detector] = None
        self._log = BackendLogger.getLogger(__name__)

        self.restore_defaults()

    def _load_from_cache(self, file: h5py.File) -> None:
        self.shared_state.config = DetectorConfig.from_json(file["config"][()])

    @is_task
    def restore_defaults(self) -> None:
        self.shared_state = SharedState(config=get_medium_range_config())
        self.broadcast()

    @is_task
    def update_sensor_id(self, *, sensor_id: int) -> None:
        self.shared_state.sensor_id = sensor_id
        self.broadcast()

    def _sync_sensor_ids(self) -> None:
        if self.client is not None:
            sensor_ids = self.client.server_info.connected_sensors

            if len(sensor_ids) > 0 and self.shared_state.sensor_id not in sensor_ids:
                self.shared_state.sensor_id = sensor_ids[0]

    @is_task
    def update_config(self, *, config: DetectorConfig) -> None:
        self.shared_state.config = config
        self.broadcast()

    def save_to_cache(self, file: h5py.File) -> None:
        _create_h5_string_dataset(file, "config", self.shared_state.config.to_json())

    @is_task
    def set_preset(self, preset_id: int) -> None:
        preset_config = self.PLUGIN_PRESETS[preset_id]
        self.shared_state.config = preset_config()
        self.broadcast()

    def load_from_record_setup(self, *, record: a121.H5Record) -> None:
        algo_group = record.get_algo_group(self.key)
        _, config, context = _load_algo_data(algo_group)
        self.shared_state.config = config
        self.shared_state.context = context
        self.shared_state.sensor_id = record.sensor_id

    def _start_session(self, recorder: Optional[a121.H5Recorder]) -> None:
        assert self.client
        self._detector_instance = Detector(
            client=self.client,
            sensor_id=self.shared_state.sensor_id,
            detector_config=self.shared_state.config,
            detector_context=self.shared_state.context,
        )
        self._detector_instance.start(recorder)
        assert self._detector_instance.detector_metadata is not None
        self.callback(
            SetupMessage(
                detector_config=self.shared_state.config,
                detector_metadata=self._detector_instance.detector_metadata,
                estimated_frame_rate=self._detector_instance.estimated_frame_rate,
            )
        )

    def end_session(self) -> None:
        if self._detector_instance is None:
            raise RuntimeError
        if self._recorder is not None:
            self._recorder.close()
        self._detector_instance.stop()

    def get_next(self) -> None:
        assert self.client
        if self._detector_instance is None:
            raise RuntimeError
        result = self._detector_instance.get_next()

        self.callback(backend.PlotMessage(result=result))


class PlotPlugin(PgPlotPlugin):
    def __init__(self, app_model: AppModel) -> None:
        super().__init__(app_model=app_model)
        self._plot_job: Optional[DetectorResult] = None
        self._is_setup = False

    def handle_message(self, message: backend.GeneralMessage) -> None:
        if isinstance(message, backend.PlotMessage):
            self._plot_job = message.result
        elif isinstance(message, SetupMessage):
            self.setup(
                detector_config=message.detector_config,
                detector_metadata=message.detector_metadata,
                estimated_frame_rate=message.estimated_frame_rate,
            )
            self._is_setup = True
        else:
            log.warn(f"{self.__class__.__name__} got an unsupported command: {message.name!r}.")

    def draw(self) -> None:
        if not self._is_setup or self._plot_job is None:
            return

        try:
            self.draw_plot_job(self._plot_job)
        finally:
            self._plot_job = None

    def setup(
        self,
        detector_config: DetectorConfig,
        detector_metadata: DetectorMetadata,
        estimated_frame_rate: float,
    ) -> None:
        self.plot_layout.clear()

        self.detector_config = detector_config
        self.distances = np.linspace(
            detector_metadata.start_m,
            detector_config.end_m,
            detector_metadata.num_points,
        )

        self.history_length_s = 5
        self.history_length_n = int(round(self.history_length_s * estimated_frame_rate))
        self.intra_history = np.zeros(self.history_length_n)
        self.inter_history = np.zeros(self.history_length_n)

        win = self.plot_layout

        self.intra_limit_lines = []
        self.inter_limit_lines = []

        # Noise estimation plot

        self.noise_plot = win.addPlot(
            row=0,
            col=0,
            title="Noise",
        )
        self.noise_plot.setMenuEnabled(False)
        self.noise_plot.setMouseEnabled(x=False, y=False)
        self.noise_plot.hideButtons()
        self.noise_plot.showGrid(x=True, y=True)
        self.noise_plot.setLabel("bottom", "Distance (m)")
        self.noise_plot.setLabel("left", "Amplitude")
        self.noise_plot.setVisible(False)
        self.noise_curve = self.noise_plot.plot(pen=et.utils.pg_pen_cycler())
        self.noise_smooth_max = et.utils.SmoothMax(self.detector_config.frame_rate)

        # Depthwise presence plot

        self.move_plot = pg.PlotItem(title="Depthwise presence")
        self.move_plot.setMenuEnabled(False)
        self.move_plot.setMouseEnabled(x=False, y=False)
        self.move_plot.hideButtons()
        self.move_plot.showGrid(x=True, y=True)
        self.move_plot.setLabel("bottom", "Distance (m)")
        self.move_plot.setLabel("left", "Norm. ampl.")
        self.move_plot.setXRange(self.distances[0], self.distances[-1])
        self.intra_curve = self.move_plot.plot(pen=et.utils.pg_pen_cycler(1))
        if not self.detector_config.intra_enable:
            self.intra_curve.hide()

        self.inter_curve = self.move_plot.plot(pen=et.utils.pg_pen_cycler(0))
        if not self.detector_config.inter_enable:
            self.inter_curve.hide()

        self.move_smooth_max = et.utils.SmoothMax(
            self.detector_config.frame_rate,
            tau_decay=1.0,
            tau_grow=0.25,
        )

        self.move_depth_line = pg.InfiniteLine(pen=pg.mkPen("k", width=1.5))
        self.move_depth_line.hide()
        self.move_plot.addItem(self.move_depth_line)

        self.present_html_format = (
            '<div style="text-align: center">'
            '<span style="color: #FFFFFF;font-size:15pt;">'
            "{}</span></div>"
        )
        not_present_html = (
            '<div style="text-align: center">'
            '<span style="color: #FFFFFF;font-size:15pt;">'
            "{}</span></div>".format("No presence detected")
        )
        self.present_text_item = pg.TextItem(
            fill=pg.mkColor(0xFF, 0x7F, 0x0E, 200),
            anchor=(0.5, 0),
        )
        self.not_present_text_item = pg.TextItem(
            html=not_present_html,
            fill=pg.mkColor(0x1F, 0x77, 0xB4, 180),
            anchor=(0.5, 0),
        )

        self.move_plot.addItem(self.present_text_item)
        self.move_plot.addItem(self.not_present_text_item)
        self.present_text_item.hide()
        self.not_present_text_item.hide()

        # Intra presence history plot

        self.intra_hist_plot = win.addPlot(
            row=1,
            col=0,
            title="Intra presence history (fast motions)",
        )
        self.intra_hist_plot.setMenuEnabled(False)
        self.intra_hist_plot.setMouseEnabled(x=False, y=False)
        self.intra_hist_plot.hideButtons()
        self.intra_hist_plot.showGrid(x=True, y=True)
        self.intra_hist_plot.setLabel("bottom", "Time (s)")
        self.intra_hist_plot.setLabel("left", "Score")
        self.intra_hist_plot.setXRange(-self.history_length_s, 0)
        self.intra_history_smooth_max = et.utils.SmoothMax(self.detector_config.frame_rate)
        self.intra_hist_plot.setYRange(0, 10)
        if not self.detector_config.intra_enable:
            intra_color = et.utils.color_cycler(1)
            intra_color = f"{intra_color}50"
            intra_dashed_pen = pg.mkPen(intra_color, width=2.5, style=QtCore.Qt.PenStyle.DashLine)
            intra_pen = pg.mkPen(intra_color, width=2)
        else:
            intra_dashed_pen = et.utils.pg_pen_cycler(1, width=2.5, style="--")
            intra_pen = et.utils.pg_pen_cycler(1)

        self.intra_hist_curve = self.intra_hist_plot.plot(pen=intra_pen)
        limit_line = pg.InfiniteLine(angle=0, pen=intra_dashed_pen)
        self.intra_hist_plot.addItem(limit_line)
        self.intra_limit_lines.append(limit_line)

        for line in self.intra_limit_lines:
            line.setPos(self.detector_config.intra_detection_threshold)

        # Inter presence history plot

        self.inter_hist_plot = win.addPlot(
            row=1,
            col=1,
            title="Inter presence history (slow motions)",
        )
        self.inter_hist_plot.setMenuEnabled(False)
        self.inter_hist_plot.setMouseEnabled(x=False, y=False)
        self.inter_hist_plot.hideButtons()
        self.inter_hist_plot.showGrid(x=True, y=True)
        self.inter_hist_plot.setLabel("bottom", "Time (s)")
        self.inter_hist_plot.setLabel("left", "Score")
        self.inter_hist_plot.setXRange(-self.history_length_s, 0)
        self.inter_history_smooth_max = et.utils.SmoothMax(self.detector_config.frame_rate)
        self.inter_hist_plot.setYRange(0, 10)
        if not self.detector_config.inter_enable:
            inter_color = et.utils.color_cycler(0)
            inter_color = f"{inter_color}50"
            inter_dashed_pen = pg.mkPen(inter_color, width=2.5, style=QtCore.Qt.PenStyle.DashLine)
            inter_pen = pg.mkPen(inter_color, width=2)
        else:
            inter_pen = et.utils.pg_pen_cycler(0)
            inter_dashed_pen = et.utils.pg_pen_cycler(0, width=2.5, style="--")

        self.inter_hist_curve = self.inter_hist_plot.plot(pen=inter_pen)
        limit_line = pg.InfiniteLine(angle=0, pen=inter_dashed_pen)
        self.inter_hist_plot.addItem(limit_line)
        self.inter_limit_lines.append(limit_line)

        for line in self.inter_limit_lines:
            line.setPos(self.detector_config.inter_detection_threshold)

        sublayout = win.addLayout(row=2, col=0, colspan=2)
        sublayout.layout.setColumnStretchFactor(0, 2)
        sublayout.addItem(self.move_plot, row=0, col=0)

    def draw_plot_job(self, data: DetectorResult) -> None:
        noise = data.processor_extra_result.lp_noise
        self.noise_curve.setData(self.distances, noise)
        self.noise_plot.setYRange(0, self.noise_smooth_max.update(noise))

        movement_x = data.presence_distance

        self.inter_curve.setData(self.distances, data.inter_depthwise_scores)
        self.intra_curve.setData(self.distances, data.intra_depthwise_scores)
        m = self.move_smooth_max.update(
            np.max(np.maximum(data.inter_depthwise_scores, data.intra_depthwise_scores))
        )
        m = max(
            m,
            2
            * np.maximum(
                self.detector_config.intra_detection_threshold,
                self.detector_config.inter_detection_threshold,
            ),
        )
        self.move_plot.setYRange(0, m)
        self.move_depth_line.setPos(movement_x)
        self.move_depth_line.setVisible(bool(data.presence_detected))

        self.set_present_text_y_pos(m)

        if data.presence_detected:
            present_text = "Presence detected at {:.0f} cm".format(movement_x * 100)
            present_html = self.present_html_format.format(present_text)
            self.present_text_item.setHtml(present_html)

            self.present_text_item.show()
            self.not_present_text_item.hide()
        else:
            self.present_text_item.hide()
            self.not_present_text_item.show()

        # Intra presence

        move_hist_xs = np.linspace(-self.history_length_s, 0, self.history_length_n)

        self.intra_history = np.roll(self.intra_history, -1)
        self.intra_history[-1] = data.intra_presence_score

        m_hist = max(
            float(np.max(self.intra_history)),
            self.detector_config.intra_detection_threshold * 1.05,
        )
        m_hist = self.intra_history_smooth_max.update(m_hist)

        self.intra_hist_plot.setYRange(0, m_hist)
        self.intra_hist_curve.setData(move_hist_xs, self.intra_history)

        # Inter presence

        self.inter_history = np.roll(self.inter_history, -1)
        self.inter_history[-1] = data.inter_presence_score

        m_hist = max(
            float(np.max(self.inter_history)),
            self.detector_config.inter_detection_threshold * 1.05,
        )
        m_hist = self.inter_history_smooth_max.update(m_hist)

        self.inter_hist_plot.setYRange(0, m_hist)
        self.inter_hist_curve.setData(move_hist_xs, self.inter_history)

    def set_present_text_y_pos(self, y: float) -> None:
        x_pos = self.distances[0] + (self.distances[-1] - self.distances[0]) / 2
        self.present_text_item.setPos(x_pos, 0.95 * y)
        self.not_present_text_item.setPos(x_pos, 0.95 * y)


class ViewPlugin(A121ViewPluginBase):
    def __init__(self, app_model: AppModel) -> None:
        super().__init__(app_model=app_model)
        self._log = logging.getLogger(__name__)

        self.start_button = QPushButton(icons.PLAY(), "Start measurement")
        self.start_button.setShortcut("space")
        self.start_button.setToolTip("Starts the session.\n\nShortcut: Space")
        self.start_button.clicked.connect(self._send_start_request)

        self.stop_button = QPushButton(icons.STOP(), "Stop")
        self.stop_button.setShortcut("space")
        self.stop_button.setToolTip("Stops the session.\n\nShortcut: Space")
        self.stop_button.clicked.connect(self._send_stop_request)

        button_group = GroupBox.grid("Controls", parent=self.sticky_widget)
        button_group.layout().addWidget(self.start_button, 0, 0)
        button_group.layout().addWidget(self.stop_button, 0, 1)

        self.misc_error_view = MiscErrorView(self.scrolly_widget)

        sensor_selection_group = GroupBox.vertical("Sensor selection", parent=self.scrolly_widget)
        self.sensor_id_pidget = pidgets.SensorIdPidgetFactory(items=[]).create(
            parent=sensor_selection_group
        )
        self.sensor_id_pidget.sig_update.connect(self._on_sensor_id_update)
        sensor_selection_group.layout().addWidget(self.sensor_id_pidget)

        self.config_editor = AttrsConfigEditor(
            title="Detector parameters",
            factory_mapping=self._get_pidget_mapping(),
            config_type=DetectorConfig,
            extra_presenter=_set_config_presenter,
            parent=self.scrolly_widget,
        )
        self.config_editor.sig_update.connect(self._on_config_update)

        goto_resource_tab_button = GotoResourceTabButton()
        goto_resource_tab_button.clicked.connect(
            lambda: app_model.sig_resource_tab_input_block_requested.emit(
                self.config_editor.get_data()
            )
        )

        sticky_layout = QVBoxLayout()
        sticky_layout.setContentsMargins(0, 0, 0, 0)

        sticky_layout.addWidget(button_group)

        scrolly_layout = QVBoxLayout()
        scrolly_layout.setContentsMargins(0, 0, 0, 0)

        self.sensor_config_status = SensorConfigEditor()
        self.sensor_config_status.set_read_only(True)
        collapsible_widget = CollapsibleWidget(
            "Current sensor settings", self.sensor_config_status, self.scrolly_widget
        )

        scrolly_layout.addWidget(goto_resource_tab_button)
        scrolly_layout.addWidget(self.misc_error_view)
        scrolly_layout.addWidget(sensor_selection_group)
        scrolly_layout.addWidget(collapsible_widget)
        scrolly_layout.addWidget(self.config_editor)

        self.sticky_widget.setLayout(sticky_layout)
        self.scrolly_widget.setLayout(scrolly_layout)

    @classmethod
    def _get_pidget_mapping(cls) -> PidgetGroupFactoryMapping:
        return _pidget_mapping.get_pidget_mapping()

    def on_backend_state_update(self, state: Optional[SharedState]) -> None:
        if state is None:
            self.config_editor.set_data(None)
            self.sensor_config_status.set_data(None)
        else:
            self.config_editor.set_data(state.config)
            self.sensor_id_pidget.set_data(state.sensor_id)

            results = state.config._collect_validation_results()

            not_handled = self.config_editor.handle_validation_results(results)

            not_handled = self.misc_error_view.handle_validation_results(not_handled)

            assert not_handled == []

            if len(results) == 0:
                sensor_config = Detector._get_sensor_config(state.config)

                self.sensor_config_status.set_data(sensor_config)

    def on_app_model_update(self, app_model: AppModel) -> None:
        self.sensor_id_pidget.set_selectable_sensors(app_model.connected_sensors)

        self.start_button.setEnabled(
            visual_policies.start_button_enabled(
                app_model, extra_condition=self.config_editor.is_ready
            )
        )
        self.stop_button.setEnabled(visual_policies.stop_button_enabled(app_model))

        visual_policies.apply_enabled_policy(
            visual_policies.config_editor_enabled,
            app_model,
            widgets=[self.config_editor, self.sensor_id_pidget],
        )

    def _on_config_update(self, config: DetectorConfig) -> None:
        BackendPlugin.update_config.rpc(self.app_model.put_task, config=config)

    def _send_defaults_request(self) -> None:
        BackendPlugin.restore_defaults.rpc(self.app_model.put_task)

    def _on_sensor_id_update(self, sensor_id: int) -> None:
        BackendPlugin.update_sensor_id.rpc(self.app_model.put_task, sensor_id=sensor_id)


def _set_config_presenter(config: Any, presentation_type: PresentationType) -> Optional[str]:
    if isinstance(config, DetectorConfig) and presentation_type is PresentationType.C_SET_CONFIG:
        step_length_call = (
            "acc_detector_presence_config_auto_step_length_set(config, true)"
            if config.step_length is None
            else f"acc_detector_presence_config_step_length_set(config, {config.step_length}U)"
        )
        profile_call = (
            "acc_detector_presence_config_auto_profile_set(config, true)"
            if config.profile is None
            else f"acc_detector_presence_config_profile_set(config, {config.profile}U)"
        )
        intra = config.intra_enable
        inter = config.inter_enable
        intra_comment = "// " if not intra else ""
        inter_comment = "// " if not inter else ""

        return f"""
static void set_config(acc_detector_presence_config_t *config, presence_preset_config_t preset)
{{
    // This snippet is generated to be compatible with RSS A121 v1.0.0
    // If there is a version missmatch the snippet might need some modification

    (void)preset;

    acc_detector_presence_config_sensor_set(config, SENSOR_ID);

    acc_detector_presence_config_start_set(config, {config.start_m:.3f}f);
    acc_detector_presence_config_end_set(config, {config.end_m:.3f}f);
    {profile_call};
    {step_length_call};
    acc_detector_presence_config_sweeps_per_frame_set(config, {config.sweeps_per_frame}U);
    acc_detector_presence_config_hwaas_set(config, {config.hwaas}U);

    acc_detector_presence_config_frame_rate_set(config, {config.frame_rate:.3f}f);
    acc_detector_presence_config_inter_frame_idle_state_set(config, ACC_CONFIG_IDLE_STATE_{config.inter_frame_idle_state.name});

    acc_detector_presence_config_intra_detection_set(config, {str(config.intra_enable).lower()});
    {intra_comment}acc_detector_presence_config_intra_detection_threshold_set(config, {config.intra_detection_threshold:.3f}f);
    {intra_comment}acc_detector_presence_config_intra_frame_time_const_set(config, {config.intra_frame_time_const:.3f}f);
    {intra_comment}acc_detector_presence_config_intra_output_time_const_set(config, {config.intra_output_time_const:.3f}f);


    acc_detector_presence_config_inter_detection_set(config, {str(config.inter_enable).lower()});
    {inter_comment}acc_detector_presence_config_inter_detection_threshold_set(config, {config.inter_detection_threshold:.3f}f);
    {inter_comment}acc_detector_presence_config_inter_frame_fast_cutoff_set(config, {config.inter_frame_fast_cutoff:.3f}f);
    {inter_comment}acc_detector_presence_config_inter_frame_slow_cutoff_set(config, {config.inter_frame_slow_cutoff:.3f}f);
    {inter_comment}acc_detector_presence_config_inter_frame_deviation_time_const_set(config, {config.inter_frame_deviation_time_const:.3f}f);
    {inter_comment}acc_detector_presence_config_inter_output_time_const_set(config, {config.inter_output_time_const:.3f});
    {inter_comment}acc_detector_presence_config_inter_frame_presence_timeout_set(config, {config.inter_frame_presence_timeout or 0.0:.3f}f);

    // This parameter is needed if the sensor is put in HIBERNATE or OFF.
    // For more information, see the Presence Detector User Guide:
    //
    // https://developer.acconeer.com
    // Documents and learning > A121 > SW > A121 Presence Detector User Guide
    //
    // acc_detector_presence_config_frame_rate_app_driven_set(config, true);
}}
"""

    return None


class PluginSpec(PluginSpecBase):
    def create_backend_plugin(
        self, callback: Callable[[Message], None], key: str
    ) -> BackendPlugin:
        return BackendPlugin(callback=callback, generation=self.generation, key=key)

    def create_view_plugin(self, app_model: AppModel) -> ViewPlugin:
        return ViewPlugin(app_model=app_model)

    def create_plot_plugin(self, app_model: AppModel) -> PlotPlugin:
        return PlotPlugin(app_model=app_model)


PRESENCE_DETECTOR_PLUGIN = PluginSpec(
    generation=PluginGeneration.A121,
    key="presence_detector",
    title="Presence detector",
    docs_link="https://docs.acconeer.com/en/latest/detectors/a121/presence_detector.html",
    description="Detect human presence.",
    family=PluginFamily.DETECTOR,
    presets=[
        PluginPresetBase(
            name="Short range",
            description="Short range",
            preset_id=PluginPresetId.SHORT_RANGE,
        ),
        PluginPresetBase(
            name="Medium range",
            description="Medium range",
            preset_id=PluginPresetId.MEDIUM_RANGE,
        ),
        PluginPresetBase(
            name="Long range",
            description="Long range",
            preset_id=PluginPresetId.LONG_RANGE,
        ),
        PluginPresetBase(
            name="Low power, Wake up",
            description="Low power presence detection, which can be used as a trigger",
            preset_id=PluginPresetId.LOW_POWER,
        ),
    ],
    default_preset_id=PluginPresetId.MEDIUM_RANGE,
)



# Copyright (c) Acconeer AB, 2023
# All rights reserved
from __future__ import annotations

import typing as t

import attributes_doc


def get_docstring(obj: t.Any) -> t.Optional[str]:
    """Gets the docstring of an object"""
    return getattr(obj, "__doc__", None)


def get_attribute_docstring(obj: t.Any, attribute_name: str) -> t.Optional[str]:
    """
    Given an object, retrieve the docstring for the attribute_name with name
    `attribute_name`.

    `attributes_doc` decorated classes are supported
    (see https://pypi.org/project/attributes-doc/)
    """
    return _get_attribute_docstring_vanilla(obj, attribute_name) or attributes_doc.get_doc(
        obj, attribute_name
    )


def _get_attribute_docstring_vanilla(obj: t.Any, attribute_name: str) -> t.Optional[str]:
    attribute = getattr(obj, attribute_name, None)
    if attribute is None:
        return None
    else:
        return get_docstring(attribute)



# Copyright (c) Acconeer AB, 2023
# All rights reserved

from __future__ import annotations

import abc
import typing as t

from .messages import Message


_ConfigT = t.TypeVar("_ConfigT", contravariant=True)


class CommunicationProtocol(abc.ABC, t.Generic[_ConfigT]):
    end_sequence: t.ClassVar[bytes] = b"\n"

    @classmethod
    @abc.abstractmethod
    def setup_command(cls, config: _ConfigT) -> bytes:
        """The `setup` command."""
        pass

    @classmethod
    @abc.abstractmethod
    def parse_message(cls, header: dict[str, t.Any], payload: bytes) -> Message:
        """Parses any supported Message given a header and a payload"""
        pass

    @classmethod
    def get_system_info_command(cls) -> bytes:
        """The `get_system_info` command."""
        return b'{"cmd":"get_system_info"}\n'

    @classmethod
    def get_sensor_info_command(cls) -> bytes:
        return b'{"cmd":"get_sensor_info"}\n'

    @classmethod
    def start_streaming_command(cls) -> bytes:
        return b'{"cmd":"start_streaming"}\n'

    @classmethod
    def stop_streaming_command(cls) -> bytes:
        return b'{"cmd":"stop_streaming"}\n'

    @classmethod
    def set_baudrate_command(cls, baudrate: int) -> bytes:
        return b'{"cmd":"set_uart_baudrate","baudrate":' + str(baudrate).encode("ascii") + b"}\n"



# Copyright (c) Acconeer AB, 2023-2024
# All rights reserved

from __future__ import annotations

import contextlib
import typing as t

import typing_extensions as te

from acconeer.exptool._core.docstrings import get_attribute_docstring, get_docstring


_T = t.TypeVar("_T")


@te.runtime_checkable
class Descriptor(te.Protocol[_T]):
    def __get__(self, obj: t.Any, objtype: t.Optional[type] = None) -> _T: ...

    def __set__(self, obj: t.Any, value: _T) -> None: ...


def delegate_field(
    delegate_name: str,
    descriptor: t.Any,
    type_: type[_T],
    *,
    doc: t.Union[str, type, None] = None,
    error_format: t.Optional[str] = None,
) -> Descriptor[_T]:
    """
    Generate a property of the outer class that modifies an inner class's fields.

    Example:

    >>> import attrs
    >>>
    >>> @attrs.mutable
    ... class Wheel:
    ...     rpm: float
    ...     angle: float
    >>>
    >>> @attrs.mutable
    ... class Bike:
    ...     front_wheel: Wheel
    ...     back_wheel: Wheel
    ...
    ...     # Note the missing type annotations here
    ...     rpm = delegate_field("back_wheel", Wheel.rpm, type_=float, doc="Rpm of the back wheel")
    ...     angle = delegate_field("front_wheel", Wheel.angle, type_=float, doc="Angle of the front wheel")
    >>>
    >>> b = Bike(Wheel(0, 0), Wheel(0, 0))
    >>> b.rpm = 70
    >>> b.angle = 45
    >>> b
    Bike(front_wheel=Wheel(rpm=0, angle=45), back_wheel=Wheel(rpm=70, angle=0))

    >>> Bike.rpm.__doc__
    'Rpm of the back wheel'
    >>> Bike.angle.__doc__
    'Angle of the front wheel'

    :param delegate_name:
            Name of the the attribute that holds an object to delegate to
    :param descriptor:
            The descriptor to delegate to.
            Works for "attrs.field"s and "property"s
    :param type_:
            The type of the resulting property.
    :param doc:
            If str:             This will be the property's docstring. An empty
                                string will set __doc__ = None.
            If a class (type):  Will attempt to find a docstring in the class,
                                using the descriptor name.
                                If no docstring can be found, an error will be raised.
            If omitted or None: Uses descriptor.__doc__ unless
                                it's None (which will result in an error)
    :param error_format:
            Used to change the AttributeError's message.
            The catched AttributeError originates from ~`getattr(self, delegate_name)`.
            It's format string that accepts the keywords
            'descriptor_name' and 'delegate_name', e.g.

                "descriptor is {descriptor_name} & delegate is {delegate_name}"

    :returns:
            A delegating property
    """
    if not isinstance(descriptor, Descriptor):
        msg = f"Passed object {descriptor} is not a descriptor"
        raise TypeError(msg)

    def getter(obj: t.Any) -> _T:
        with _more_helpful_attribute_error(error_format, delegate_name, descriptor):
            delegate = getattr(obj, delegate_name)

        return descriptor.__get__(delegate)  # type: ignore[no-any-return]

    def setter(obj: t.Any, value: _T) -> None:
        with _more_helpful_attribute_error(error_format, delegate_name, descriptor):
            delegate = getattr(obj, delegate_name)

        descriptor.__set__(delegate, value)

    return property(
        fget=_with_return_type_annotation(getter, type_),
        fset=setter,
        doc=_determine_docstring(doc, descriptor),
    )


def _descriptor_given_name(descriptor: t.Any) -> t.Optional[str]:
    if isinstance(descriptor, property):
        return getattr(descriptor.fget, "__name__", None)

    return getattr(descriptor, "__name__", None)


@contextlib.contextmanager
def _more_helpful_attribute_error(
    error_format: t.Optional[str],
    delegate_name: str,
    descriptor: t.Any,
) -> t.Iterator[None]:
    try:
        yield
    except AttributeError as e:
        if error_format is None:
            raise e

        raise AttributeError(
            error_format.format(
                descriptor_name=_descriptor_given_name(descriptor),
                delegate_name=delegate_name,
            )
        ) from None


def _with_return_type_annotation(
    f: t.Callable[..., _T], return_type: type[_T]
) -> t.Callable[..., _T]:
    """
    This function patches a function's return type annotation.
    If used on the getter of a property, Sphinx seems to collect
    that type as the property return type
    """
    f.__annotations__["return"] = return_type.__name__
    return f


def _determine_docstring(
    doc: t.Union[str, type, None], descriptor: Descriptor[t.Any]
) -> t.Optional[str]:
    """See behaviour of "doc" parameter in delegate_field."""
    if isinstance(doc, str):
        return None if doc == "" else doc
    elif doc is None and get_docstring(descriptor) is not None:
        return get_docstring(descriptor)
    elif isinstance(doc, type):
        descriptor_name = _descriptor_given_name(descriptor)
        if descriptor_name is not None:
            docstring = get_attribute_docstring(doc, descriptor_name)
            if docstring is not None:
                return docstring

    raise RuntimeError(
        f"Could not automatically find docstring for {descriptor}. "
        + 'Specify docs manually (doc="...") or skip it (doc="")'
    )



# Copyright (c) Acconeer AB, 2023-2024
# All rights reserved

from __future__ import annotations

import attrs
from serial.serialutil import SerialException

import acconeer.exptool
from acconeer.exptool._core.communication.comm_devices import (
    CommDeviceError,
    get_one_serial_device,
    get_one_usb_device,
)
from acconeer.exptool._core.entities import (
    ClientInfo,
    SerialInfo,
    USBInfo,
)

from . import (
    BufferedLink,
    ExploreSerialLink,
    LinkError,
    NullLink,
    NullLinkError,
    SocketLink,
    USBLink,
)


def link_factory(client_info: ClientInfo) -> BufferedLink:
    if client_info.socket is not None:
        return SocketLink(host=client_info.socket.ip_address, port=client_info.socket.tcp_port)

    if client_info.serial is not None:
        return ExploreSerialLink(
            port=client_info.serial.port,
            flowcontrol=client_info.serial.flow_control,
        )

    if client_info.usb is not None:
        usb_device = None
        if client_info.usb.serial_number is not None:
            usb_device = (
                acconeer.exptool._core.communication.comm_devices.get_usb_device_by_serial(
                    client_info.usb.serial_number, only_accessible=False
                )
            )
        else:
            usb_device = get_one_usb_device()

        if usb_device is not None:
            return USBLink(
                vid=usb_device.vid,
                pid=usb_device.pid,
                serial=usb_device.serial,
            )

    return NullLink()


def autodetermine_client_info(client_info: ClientInfo) -> ClientInfo:
    error_message = ""
    try:
        usb_info = client_info.usb
        if usb_info is None:
            usb_device = get_one_usb_device(only_accessible=True)
            usb_info = USBInfo(
                vid=usb_device.vid, pid=usb_device.pid, serial_number=usb_device.serial
            )
        return attrs.evolve(
            client_info,
            usb=usb_info,
        )

    except CommDeviceError as exc:
        error_message += f"\nUSB: {str(exc)}"
        pass

    try:
        serial_info = client_info.serial
        if serial_info is None:
            serial_info = SerialInfo(port=str(get_one_serial_device().port))
        return attrs.evolve(client_info, serial=serial_info)
    except CommDeviceError as exc:
        error_message += f"\nSerial: {str(exc)}"

    msg = f"Cannot auto detect:{error_message}"
    raise LinkError(msg)


def ensure_connected_link(client_info: ClientInfo) -> tuple[BufferedLink, ClientInfo]:
    """
    Tries connecting a link defined by the given ClientInfo
    before trying to autodetermine a link which can be used.

    Returns the connected link and a new ClientInfo that describes the connected link.
    """
    link = link_factory(client_info)

    try:
        link.connect()
        return link, client_info
    except NullLinkError:
        new_client_info = autodetermine_client_info(client_info)
        return ensure_connected_link(new_client_info)
    except SerialException as exc:
        if "Permission denied" in str(exc):
            text = "\n".join(
                [
                    "You are probably missing permissions to access the serial port.",
                    "",
                    "Run the setup script to fix it:",
                    "$ python -m acconeer.exptool.setup",
                    "",
                    "Reboot for the changes to take effect.",
                ]
            )
            raise LinkError(text) from exc
        else:
            raise



# Copyright (c) Acconeer AB, 2023-2024
# All rights reserved

from __future__ import annotations

import typing as t
from typing import Callable, Optional

import attrs
import numpy as np
from numpy import typing as npt


_S = t.TypeVar("_S")
_T = t.TypeVar("_T")
_U = t.TypeVar("_U")


def ndarray_isclose(a: npt.ArrayLike, b: npt.ArrayLike) -> bool:
    return bool(np.isclose(a, b, equal_nan=True).all())


def _dict_wrapper(f: Callable[[_S, _T], bool]) -> Callable[[dict[_U, _S], dict[_U, _T]], bool]:
    def wrapper(a: dict[_U, _S], b: dict[_U, _T]) -> bool:
        if a.keys() != b.keys():
            return False

        return all(f(a[key], b[key]) for key in a)

    return wrapper


def _optional_wrapper(f: Callable[[_S, _T], bool]) -> Callable[[Optional[_S], Optional[_T]], bool]:
    def wrapper(a: Optional[_S], b: Optional[_T]) -> bool:
        if a is None or b is None:
            return a is b
        else:
            return f(a, b)

    return wrapper


attrs_ndarray_eq = attrs.cmp_using(eq=np.array_equal)
attrs_optional_ndarray_eq = attrs.cmp_using(eq=_optional_wrapper(np.array_equal))
attrs_ndarray_isclose = attrs.cmp_using(eq=ndarray_isclose)
attrs_optional_ndarray_isclose = attrs.cmp_using(eq=_optional_wrapper(ndarray_isclose))
attrs_dict_ndarray_isclose = attrs.cmp_using(eq=_dict_wrapper(ndarray_isclose))


def attrs_ndarray_validator(
    ndim: int, element_type: np.dtype[np.generic]
) -> t.Callable[[t.Any, t.Any, t.Any], None]:
    """
    Creates a validator to be used in attrs fields. Does the following checks:

    1. The new value is a np.ndarray
    2. It has ``ndim`` dimensions
    3. Its elements have the type ``element_type`` (e.g. np.float64, np.int_, etc.)
    """

    def inner(_: t.Any, attribute: attrs.Attribute[t.Any], value: t.Any) -> None:
        if not isinstance(value, np.ndarray):
            msg = f"{attribute.name!r} needs to be a numpy array"
            raise TypeError(msg)

        if value.ndim != ndim:
            raise ValueError(
                f"{attribute.name!r} has {value.ndim} dimensions. "
                + f"It needs to have {ndim} dimensions"
            )

        if value.dtype != element_type:
            msg = f"{attribute.name!r} needs to have dtype {element_type!r}"
            raise ValueError(msg)

    return inner



# Copyright (c) Acconeer AB, 2023-2025
# All rights reserved

from __future__ import annotations

from typing import Optional, Tuple

import attrs
import numpy as np
import numpy.typing as npt
import scipy
from scipy.signal import welch

from acconeer.exptool import a121
from acconeer.exptool._core.class_creation.attrs import attrs_ndarray_isclose
from acconeer.exptool.a121.algo import (
    AlgoProcessorConfigBase,
    ProcessorBase,
    double_buffering_frame_filter,
)
from acconeer.exptool.a121.algo._utils import (
    APPROX_BASE_STEP_LENGTH_M,
    PERCEIVED_WAVELENGTH,
    find_peaks,
    get_distances_m,
)


@attrs.mutable(kw_only=True)
class ProcessorConfig(AlgoProcessorConfigBase):
    surface_distance: float = attrs.field(default=1)
    time_series_length: int = attrs.field(default=512)
    slow_zone: int = attrs.field(default=3)
    psd_lp_coeff: float = attrs.field(default=0.75)
    cfar_guard: int = attrs.field(default=6)
    cfar_win: int = attrs.field(default=6)
    cfar_sensitivity: float = attrs.field(default=0.15)
    velocity_lp_coeff: float = attrs.field(default=0.95)
    max_peak_interval_s: float = attrs.field(default=4)

    def _collect_validation_results(
        self, config: a121.SessionConfig
    ) -> list[a121.ValidationResult]:
        validation_results: list[a121.ValidationResult] = []

        if len(config.sensor_config.subsweeps) > 1:
            validation_results.append(
                a121.ValidationError(
                    config.sensor_config,
                    "subsweeps",
                    "Multiple subsweeps are not supported",
                )
            )

        if config.sensor_config.sweep_rate is None:
            validation_results.append(
                a121.ValidationError(
                    config.sensor_config,
                    "sweep_rate",
                    "Must be set",
                )
            )

        if config.sensor_config.start_point * APPROX_BASE_STEP_LENGTH_M <= self.surface_distance:
            validation_results.append(
                a121.ValidationError(
                    self,
                    "start_point",
                    "Start point must be > surface distance",
                )
            )

        return validation_results


@attrs.frozen(kw_only=True)
class ProcessorContext: ...


@attrs.frozen(kw_only=True)
class ProcessorExtraResult:
    """
    Contains information for visualization in ET.
    """

    max_bin_vertical_vs: npt.NDArray[np.float64] = attrs.field(eq=attrs_ndarray_isclose)
    peak_width: float = attrs.field()

    vertical_velocities: npt.NDArray[np.float64] = attrs.field(eq=attrs_ndarray_isclose)
    psd: npt.NDArray[np.float64] = attrs.field(eq=attrs_ndarray_isclose)
    peak_idx: Optional[np.int_] = attrs.field()
    psd_threshold: npt.NDArray[np.float64] = attrs.field(eq=attrs_ndarray_isclose)


@attrs.frozen(kw_only=True)
class ProcessorResult:
    estimated_v: float = attrs.field()
    distance_m: float = attrs.field()

    extra_result: ProcessorExtraResult = attrs.field()


class Processor(ProcessorBase[ProcessorResult]):
    MIN_PEAK_VS = 0.1

    def __init__(
        self,
        *,
        sensor_config: a121.SensorConfig,
        metadata: a121.Metadata,
        processor_config: ProcessorConfig,
        context: Optional[ProcessorContext] = None,
    ) -> None:
        self.sensor_config = sensor_config
        self.metadata = metadata
        self.processor_config = processor_config

        # Will never happen because checked in _collect_validation_results
        assert self.sensor_config.sweep_rate is not None

        self.sweep_rate = self.sensor_config.sweep_rate

        self.distances = get_distances_m(self.sensor_config, metadata)
        self.num_distances = self.distances.size
        self.sweeps_per_frame = self.sensor_config.sweeps_per_frame

        if not sensor_config.continuous_sweep_mode:
            self.time_series_length = self.sweeps_per_frame
        else:
            if sensor_config.sweeps_per_frame > processor_config.time_series_length:
                msg = "time_series_length must be >= sweeps_per_frame"
                raise ValueError(msg)

            self.time_series_length = processor_config.time_series_length

        self.time_series = np.zeros(
            [self.time_series_length, self.num_distances], dtype=np.complex128
        )

        self.surface_distance = processor_config.surface_distance

        if sensor_config.frame_rate is None:
            estimated_frame_rate = self.sweep_rate / self.sweeps_per_frame
        else:
            estimated_frame_rate = sensor_config.frame_rate

        self.max_peak_interval_n = processor_config.max_peak_interval_s * estimated_frame_rate

        # welch
        self.update_index = 0
        self.segment_length = self.time_series_length // 4
        if np.mod(self.segment_length, 2) != 0:
            self.segment_length += 1

        self.middle_idx = int(np.around(self.segment_length / 2))

        _, bin_fs = self.scipy_welch(self.time_series, self.sweep_rate)
        self.bin_rad_vs = bin_fs * PERCEIVED_WAVELENGTH

        self.max_bin_vertical_vs = self.bin_rad_vs * self.get_angle_correction(self.distances[0])

        self.lp_velocity = 0.0

        self.lp_psds = np.zeros([self.segment_length, self.num_distances])
        self.psd_lp_coeff = processor_config.psd_lp_coeff

        # cfar
        self.cfar_guard_length = processor_config.cfar_guard
        self.cfar_win_length = processor_config.cfar_win

        if not processor_config.cfar_sensitivity > 0:
            msg = "cfar_sensitivity must be > 0"
            raise ValueError(msg)

        self.cfar_sensitivity = processor_config.cfar_sensitivity

        self.slow_zone = processor_config.slow_zone
        self.velocity_lp_coeff = processor_config.velocity_lp_coeff

        self.wait_n = 0

    @staticmethod
    def _dynamic_sf(static_sf: float, update_index: int) -> float:
        return min(static_sf, 1.0 - 1.0 / (1.0 + update_index))

    def scipy_welch(
        self, sweeps: npt.NDArray[np.complex128], sweep_rate: float
    ) -> Tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
        psds = []
        for i in np.arange(self.num_distances):
            freqs, psd = welch(
                x=sweeps[:, i],
                fs=sweep_rate,
                window="hann",
                nperseg=self.segment_length,
                noverlap=0,
                average="mean",
                axis=0,
                return_onesided=False,
            )
            freqs = scipy.fft.fftshift(freqs)
            psd = scipy.fft.fftshift(psd)
            psds.append(psd)

        return np.array(psds).T, freqs

    def get_angle_correction(self, distance: float) -> float:
        # distanca > self.surface_distance is checked in sensor config
        insonation_angle = np.arcsin(self.surface_distance / distance)
        angle_correction = 1 / np.cos(insonation_angle)

        return angle_correction  # type: ignore[no-any-return]

    def get_distance_idx(self, psds: npt.NDArray[np.float64]) -> int:
        max_negative_side = np.max(psds[self.middle_idx + self.slow_zone :, :], axis=0)
        max_positive_side = np.max(psds[: self.middle_idx - self.slow_zone, :], axis=0)
        max_amps = np.maximum(max_negative_side, max_positive_side)

        return np.argmax(max_amps, axis=0)  # type: ignore[no-any-return]

    def get_cfar_threshold(self, psd: npt.NDArray[np.float64]) -> npt.NDArray[np.float64]:
        window_length = self.cfar_win_length
        guard_length = self.cfar_guard_length
        margin = window_length + guard_length
        half_sweep_len_without_margin = int(np.around(psd.shape[0] / 2 - margin))
        min_psd = np.min(psd)

        # One-sided CFAR
        threshold_cfar = np.full(psd.shape, np.nan)
        filt_psd = np.convolve(psd, np.ones(window_length), "valid") / window_length
        threshold_cfar[margin : self.middle_idx] = filt_psd[:half_sweep_len_without_margin]
        threshold_cfar[self.middle_idx : -margin] = filt_psd[-half_sweep_len_without_margin:]

        threshold_cfar[:margin] = threshold_cfar[margin]
        threshold_cfar[-margin:] = threshold_cfar[-margin - 1]

        threshold_cfar += min_psd

        return threshold_cfar * 1 / self.cfar_sensitivity

    def cfar_peaks(
        self, threshold: npt.NDArray[np.float64], psd: npt.NDArray[np.float64]
    ) -> list[int]:
        return find_peaks(psd, threshold)

    @staticmethod
    def _merge_peaks(
        min_peak_to_peak_vel: float,
        velocities: npt.NDArray[np.float64],
        energies: npt.NDArray[np.float64],
    ) -> Tuple[npt.NDArray[np.float64], npt.NDArray[np.float64], npt.NDArray[np.float64]]:
        sorting_order = np.argsort(velocities)
        velocities_sorted = velocities[sorting_order]
        energies_sorted = energies[sorting_order]

        peak_cluster_idxs = np.where(min_peak_to_peak_vel < np.diff(velocities_sorted))[0] + 1
        velocities_merged = [
            np.mean(cluster)
            for cluster in np.split(velocities_sorted, peak_cluster_idxs)
            if velocities.size != 0
        ]

        velocities_width = [
            np.abs(np.max(cluster) - np.min(cluster))
            for cluster in np.split(velocities_sorted, peak_cluster_idxs)
            if velocities.size != 0
        ]

        energies_merged = [
            np.mean(cluster)
            for cluster in np.split(energies_sorted, peak_cluster_idxs)
            if velocities.size != 0
        ]

        return np.array(velocities_merged), np.array(velocities_width), np.array(energies_merged)

    def _get_peak_velocity(
        self,
        velocities: npt.NDArray[np.float64],
        energies: npt.NDArray[np.float64],
        bin_vs: npt.NDArray[np.float64],
    ) -> float:
        idxs = energies.argsort()
        slow_vs = []
        valid_vs = []
        for i in idxs:
            if np.abs(velocities[i]) < bin_vs[self.middle_idx + self.slow_zone]:
                slow_vs.append(velocities[i])
            else:
                valid_vs.append(velocities[i])

        if len(valid_vs) > 0:
            return valid_vs[-1]  # type: ignore[no-any-return]
        else:
            return slow_vs[-1]  # type: ignore[no-any-return]

    def get_velocity_estimate(
        self,
        bin_vertical_vs: npt.NDArray[np.float64],
        peak_idxs: list[int],
        psd: npt.NDArray[np.float64],
    ) -> Tuple[float, np.int_, float]:
        velocities, peak_widths, energies = self._merge_peaks(
            self.MIN_PEAK_VS, bin_vertical_vs[peak_idxs], psd[peak_idxs]
        )
        vertical_v = self._get_peak_velocity(velocities, energies, bin_vertical_vs)
        peak_width = peak_widths[np.nonzero(velocities == vertical_v)]
        peak_idx = np.argmin(np.abs(bin_vertical_vs - vertical_v))

        return vertical_v, peak_idx, peak_width[0]

    def get_velocity_estimate_slow_zone(
        self,
        bin_vertical_vs: npt.NDArray[np.float64],
        peak_idxs: list[int],
        psd: npt.NDArray[np.float64],
    ) -> Tuple[float, np.int_]:
        velocities = bin_vertical_vs[peak_idxs]
        energies = psd[peak_idxs]
        vertical_v = self._get_peak_velocity(velocities, energies, bin_vertical_vs)
        peak_idx = np.argmin(np.abs(bin_vertical_vs - vertical_v))

        return vertical_v, peak_idx

    def process(self, result: a121.Result) -> ProcessorResult:
        data_segment = double_buffering_frame_filter(result._frame)

        self.time_series = np.roll(self.time_series, axis=0, shift=-self.sweeps_per_frame)
        self.time_series[-self.sweeps_per_frame :, :] = data_segment

        psds, _ = self.scipy_welch(self.time_series, self.sweep_rate)
        if self.update_index * self.sweeps_per_frame < self.time_series_length:
            self.lp_psds = psds

        self.lp_psds = self.lp_psds * self.psd_lp_coeff + psds * (1 - self.psd_lp_coeff)

        distance_idx = self.get_distance_idx(self.lp_psds)
        self.distance_idx = distance_idx
        distance = self.distances[distance_idx]
        psd = self.lp_psds[:, distance_idx]
        bin_vertical_vs = self.bin_rad_vs * self.get_angle_correction(distance)

        psd_cfar = self.get_cfar_threshold(psd)
        psd_peak_idxs = self.cfar_peaks(psd_cfar, psd)

        if len(psd_peak_idxs) > 0:
            if (
                np.max(np.abs(bin_vertical_vs[psd_peak_idxs]))
                > bin_vertical_vs[self.middle_idx + self.slow_zone]
            ):
                vertical_v, peak_idx, peak_width = self.get_velocity_estimate(
                    bin_vertical_vs, psd_peak_idxs, psd
                )
            else:
                vertical_v, peak_idx = self.get_velocity_estimate_slow_zone(
                    bin_vertical_vs, psd_peak_idxs, psd
                )
                peak_width = 0

            if np.abs(self.lp_velocity) > 0 and vertical_v / self.lp_velocity < 0.8:
                if self.wait_n < self.max_peak_interval_n:
                    vertical_v = self.lp_velocity
                    self.wait_n += 1
            else:
                self.wait_n = 0

        else:
            if self.wait_n < self.max_peak_interval_n:
                vertical_v = self.lp_velocity
                self.wait_n += 1
            else:
                vertical_v = 0

            peak_idx = None
            peak_width = 0

        sf = self._dynamic_sf(self.velocity_lp_coeff, self.update_index)
        if (self.update_index * self.sweeps_per_frame) > self.time_series_length:
            self.lp_velocity = sf * self.lp_velocity + (1 - sf) * vertical_v

        self.update_index += 1

        extra_result = ProcessorExtraResult(
            max_bin_vertical_vs=self.max_bin_vertical_vs,
            peak_width=peak_width,
            vertical_velocities=bin_vertical_vs,
            psd=psd,
            peak_idx=peak_idx,
            psd_threshold=psd_cfar,
        )

        return ProcessorResult(
            estimated_v=self.lp_velocity,
            distance_m=distance,
            extra_result=extra_result,
        )



# Copyright (c) Acconeer AB, 2024
# All rights reserved

# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from __future__ import annotations

import typing as t

from PySide6.QtCore import QPoint, QRect, QSize
from PySide6.QtGui import QResizeEvent
from PySide6.QtWidgets import QVBoxLayout, QWidget


class Overlay(QWidget):
    """Implements overlaying a number of widgets on top of another widget

    'base' is the background widget and 'overlays' is pairs of
    - 'overlay_widget': The actual widget
    - 'positioner': A function that calculates the position of the widget

    'positioner' is a callable that accepts
    - the '.rect()' of 'base' (left, top, width, height)
    - the '.sizeHint()' of 'overlay_widget' (preffered width, height)
    and returns the new rect (left, top, width, height) of 'overlay_widget' should have
    """

    @staticmethod
    def positioner_top_right(base_rect: QRect, overlay_size: QSize) -> QRect:
        top_left = base_rect.topRight() - QPoint(overlay_size.width() + 7, 0)
        return QRect(top_left, overlay_size)

    def __init__(
        self,
        base: QWidget,
        overlays: list[tuple[QWidget, t.Callable[[QRect, QSize], QRect]]],
        parent: t.Optional[QWidget] = None,
    ) -> None:
        super().__init__(parent=parent)

        self._base = base
        self._overlays = overlays

        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self._base)
        self.setLayout(layout)

        for widget, _ in self._overlays:
            widget.setParent(self)

    def resizeEvent(self, event: QResizeEvent) -> None:
        super().resizeEvent(event)
        for widget, positioner in self._overlays:
            positioned = positioner(self._base.rect(), widget.sizeHint())
            widget.setGeometry(positioned)



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved
from __future__ import annotations

import typing as t

import typing_extensions as te

from .message import Message, ParseError


class StartStreamingResponseHeader(te.TypedDict):
    status: te.Literal["start"]


class StopStreamingResponseHeader(te.TypedDict):
    status: te.Literal["stop"]


class StartStreamingResponse(Message):
    @classmethod
    def parse(cls, header: t.Dict[str, t.Any], payload: bytes) -> StartStreamingResponse:
        t.cast(StartStreamingResponseHeader, header)

        if header["status"] == "start":
            return cls()
        else:
            raise ParseError


class StopStreamingResponse(Message):
    @classmethod
    def parse(cls, header: t.Dict[str, t.Any], payload: bytes) -> StopStreamingResponse:
        t.cast(StopStreamingResponseHeader, header)

        if header["status"] == "stop":
            return cls()
        else:
            raise ParseError



# Copyright (c) Acconeer AB, 2025
# All rights reserved
from __future__ import annotations

import typing as t
from enum import Flag, auto

import numpy as np
import typing_extensions as te

from acconeer.exptool import a121
from acconeer.exptool.a121.algo import (
    APPROX_BASE_STEP_LENGTH_M,
    ENVELOPE_FWHM_M,
    RLG_PER_HWAAS_MAP,
    calc_processing_gain,
    get_distance_filter_edge_margin,
    select_prf_m,
)

from ._aggregator import ProcessorSpec
from ._processors import MeasurementType, Processor, ProcessorConfig, ThresholdMethod


if t.TYPE_CHECKING:
    from ._detector import DetectorConfig

_MIN_LEAKAGE_FREE_DIST_M = {
    a121.Profile.PROFILE_1: 0.12,
    a121.Profile.PROFILE_2: 0.28,
    a121.Profile.PROFILE_3: 0.56,
    a121.Profile.PROFILE_4: 0.76,
    a121.Profile.PROFILE_5: 1.28,
}
_MIN_NUM_POINTS_IN_ENVELOPE_FWHM_SPAN = 4.0
_VALID_STEP_LENGTHS = [1, 2, 3, 4, 6, 8, 12, 24]
_NUM_SUBSWEEPS_IN_SENSOR_CONFIG = 4
_HWAAS_MIN_DISTANCE = 1.0
_MAX_HWAAS = 511
_MIN_HWAAS = 1


class RangeTypes(Flag):
    CLOSE_RANGE = auto()
    TRANSITION_P1 = auto()
    TRANSITION_P3 = auto()
    MAX_PROFILE = auto()

    # Aliases
    ANY_TRANSITION = TRANSITION_P1 | TRANSITION_P3
    FAR_RANGE = TRANSITION_P1 | TRANSITION_P3 | MAX_PROFILE

    @property
    def closest_range_type(self) -> te.Self:
        (closest, *_) = self._masked_in_members
        return closest

    # These features where added in Python 3.11, so can't rely
    # on those yet.
    def __len__(self) -> int:
        """Basically a "bit" count."""
        return len(self._masked_in_members)

    def __iter__(self) -> t.Iterator[te.Self]:
        """Iterate over set "bits"."""
        return iter(self._masked_in_members)

    @property
    def _masked_in_members(self) -> list[te.Self]:
        return [
            mem  # type: ignore[misc]
            for mem in [self.CLOSE_RANGE, self.TRANSITION_P1, self.TRANSITION_P3, self.MAX_PROFILE]
            if bool(self & mem)  # type: ignore[operator]
        ]


def _clamp(value: float, min_val: float, max_val: float) -> float:
    return min(max(value, min_val), max_val)


def _range_overlap(a: tuple[float, float], b: tuple[float, float]) -> t.Optional[float]:
    (a_start, a_end) = sorted(a)
    (b_start, b_end) = sorted(b)
    overlap_start = _clamp(a_start, b_start, b_end)
    overlap_end = _clamp(a_end, b_start, b_end)
    ranges_overlapping = not (a_start >= b_end or b_start >= a_end)
    return abs(overlap_end - overlap_start) if ranges_overlapping else None


def _calc_hwaas(
    profile: a121.Profile,
    step_length: int,
    distance_m: float,
    signal_quality: float,
    reflector_shape_exponent: float,
) -> int:
    rlg_per_hwaas = RLG_PER_HWAAS_MAP[profile]

    processing_gain = calc_processing_gain(profile, step_length)
    subsweep_end_point_m = max(distance_m, _HWAAS_MIN_DISTANCE)
    rlg = (
        signal_quality
        + reflector_shape_exponent * 10 * np.log10(subsweep_end_point_m)
        - np.log10(processing_gain)
    )
    hwaas = int(round(10 ** ((rlg - rlg_per_hwaas) / 10)))
    if hwaas < _MIN_HWAAS:
        return _MIN_HWAAS
    elif hwaas > _MAX_HWAAS:
        return _MAX_HWAAS
    else:
        return hwaas


def _get_range_types(config: DetectorConfig) -> RangeTypes:
    min_dist_m = _calc_leakage_free_min_dist(config)
    configured_range = (config.start_m, config.end_m)

    range_types = RangeTypes(0)

    measuring_in_direct_leakage = config.start_m < min_dist_m[a121.Profile.PROFILE_1]
    if measuring_in_direct_leakage:
        if config.close_range_leakage_cancellation:
            range_types |= RangeTypes.CLOSE_RANGE
        elif config.max_profile is a121.Profile.PROFILE_1:
            range_types |= RangeTypes.MAX_PROFILE
        else:
            range_types |= RangeTypes.TRANSITION_P1

    if config.max_profile.value > a121.Profile.PROFILE_1.value:
        bounding_profile = min(a121.Profile.PROFILE_3, config.max_profile, key=lambda p: p.value)
        transition_p1_range = (min_dist_m[a121.Profile.PROFILE_1], min_dist_m[bounding_profile])

        step_length_m = (
            _limit_step_length(a121.Profile.PROFILE_1, config.max_step_length)
            * APPROX_BASE_STEP_LENGTH_M
        )

        min_overlap = min(step_length_m, config.end_m - config.start_m)
        range_overlap = _range_overlap(transition_p1_range, configured_range)

        if range_overlap is not None and range_overlap >= min_overlap:
            range_types |= RangeTypes.TRANSITION_P1

    if config.max_profile.value > a121.Profile.PROFILE_3.value:
        transition_p3_range = (min_dist_m[a121.Profile.PROFILE_3], min_dist_m[config.max_profile])

        step_length_m = (
            _limit_step_length(a121.Profile.PROFILE_3, config.max_step_length)
            * APPROX_BASE_STEP_LENGTH_M
        )

        min_overlap = min(step_length_m, config.end_m - config.start_m)
        range_overlap = _range_overlap(transition_p3_range, configured_range)

        if range_overlap is not None and range_overlap >= min_overlap:
            range_types |= RangeTypes.TRANSITION_P3

    if min_dist_m[config.max_profile] < config.end_m:
        range_types |= RangeTypes.MAX_PROFILE

    return range_types


def _cfar_margins(profile: a121.Profile, step_length: int) -> int:
    cfar_margin = Processor.calc_cfar_margin(profile, step_length) * step_length
    return cfar_margin


def _close_range_margins(step_length: int, range_types: RangeTypes) -> tuple[int, int]:
    filter_margin_p = (
        get_distance_filter_edge_margin(a121.Profile.PROFILE_1, step_length) * step_length
    )
    left_margin = filter_margin_p

    if range_types & RangeTypes.ANY_TRANSITION:
        right_margin = 2 * filter_margin_p
    else:
        right_margin = filter_margin_p

    return (left_margin, right_margin)


def _transition_p1_margins(step_length: int, range_types: RangeTypes) -> tuple[int, int]:
    filter_margin_p = (
        get_distance_filter_edge_margin(a121.Profile.PROFILE_1, step_length) * step_length
    )
    right_margin = filter_margin_p

    if range_types & RangeTypes.CLOSE_RANGE:
        left_margin = 2 * filter_margin_p
    else:
        left_margin = filter_margin_p

    return (left_margin, right_margin)


def _transition_p3_margins(step_length: int, range_types: RangeTypes) -> tuple[int, int]:
    filter_margin_p = (
        get_distance_filter_edge_margin(a121.Profile.PROFILE_3, step_length) * step_length
    )
    right_margin = filter_margin_p

    if range_types & (RangeTypes.TRANSITION_P1 | RangeTypes.CLOSE_RANGE):
        left_margin = 2 * filter_margin_p
    else:
        left_margin = filter_margin_p

    return (left_margin, right_margin)


def _max_profile_margins(
    step_length: int,
    max_profile: a121.Profile,
    range_types: RangeTypes,
) -> tuple[int, int]:
    filter_margin_p = get_distance_filter_edge_margin(max_profile, step_length) * step_length
    right_margin = filter_margin_p

    if range_types & ~RangeTypes.MAX_PROFILE:
        left_margin = 2 * filter_margin_p
    else:
        left_margin = filter_margin_p

    return (left_margin, right_margin)


def detector_config_to_processor_specs(
    config: DetectorConfig,
    sensor_ids: list[int],
    num_far_subsweeps: int,
) -> list[ProcessorSpec]:
    if num_far_subsweeps > 4 or num_far_subsweeps < 0:
        raise ValueError

    far_range_processor_config = ProcessorConfig(
        threshold_method=config.threshold_method,
        fixed_threshold_value=config.fixed_threshold_value,
        fixed_strength_threshold_value=config.fixed_strength_threshold_value,
        threshold_sensitivity=config.threshold_sensitivity,
        reflector_shape=config.reflector_shape,
    )
    close_range_processor_config = ProcessorConfig(
        threshold_method=ThresholdMethod.RECORDED,
        measurement_type=MeasurementType.CLOSE_RANGE,
        threshold_sensitivity=config.threshold_sensitivity,
        reflector_shape=config.reflector_shape,
    )

    range_types = _get_range_types(config)
    processor_specs = []
    if range_types & RangeTypes.CLOSE_RANGE:
        cr_proc_spec = ProcessorSpec(
            processor_config=close_range_processor_config,
            group_index=0,
            subsweep_indexes=[0, 1],
        )
        processor_specs.append(cr_proc_spec)

    # Iterator of available subsweep indexes. Will be consumed
    # in the for-loop below.
    far_subsweep_idx_iter = iter(range(num_far_subsweeps))
    far_group_idx = 1 if RangeTypes.CLOSE_RANGE in range_types else 0

    for range_type in range_types & RangeTypes.FAR_RANGE:
        if range_type & RangeTypes.ANY_TRANSITION:
            # If "Transition": Use a single subsweep
            transition_subsweep_idx = next(far_subsweep_idx_iter)

            transition_proc_spec = ProcessorSpec(
                processor_config=far_range_processor_config,
                group_index=far_group_idx,
                subsweep_indexes=[transition_subsweep_idx],
            )

            processor_specs.append(transition_proc_spec)
        elif range_type & RangeTypes.MAX_PROFILE:
            # If "Max Profile": Use the rest of the available subsweeps
            max_profile_subsweep_idxs = list(far_subsweep_idx_iter)

            max_profile_proc_spec = ProcessorSpec(
                processor_config=far_range_processor_config,
                group_index=far_group_idx,
                subsweep_indexes=max_profile_subsweep_idxs,
            )

            processor_specs.append(max_profile_proc_spec)

    return processor_specs


def detector_config_to_session_config(
    config: DetectorConfig,
    sensor_ids: list[int],
) -> a121.SessionConfig:
    groups = []

    range_types = _get_range_types(config)
    prf = _get_max_prf(config)

    min_dist_m = _calc_leakage_free_min_dist(config)

    # CLOSE RANGE
    if RangeTypes.CLOSE_RANGE & range_types:
        start_m = config.start_m
        end_m = min(config.end_m, min_dist_m[a121.Profile.PROFILE_1])

        step_length = _limit_step_length(a121.Profile.PROFILE_1, config.max_step_length)
        (left_filter_margin, right_filter_margin) = _close_range_margins(step_length, range_types)
        marginless_end_point = _m_to_point(start_m, end_m, step_length)
        end_point = marginless_end_point + right_filter_margin
        hwaas = _calc_hwaas(
            a121.Profile.PROFILE_1,
            step_length,
            end_point * APPROX_BASE_STEP_LENGTH_M,
            config.signal_quality,
            config.reflector_shape.exponent,
        )
        marginless_start_point = _m_to_point(
            config.start_m,
            config.start_m,
            step_length,
        )
        start_point = marginless_start_point - left_filter_margin
        num_points = int((end_point - start_point) / step_length)

        loopback_subsweep = a121.SubsweepConfig(
            start_point=0,
            num_points=1,
            step_length=1,
            profile=a121.Profile.PROFILE_4,
            hwaas=hwaas,
            receiver_gain=15,
            phase_enhancement=True,
            iq_imbalance_compensation=True,
            enable_loopback=True,
            prf=a121.PRF.PRF_15_6_MHz,
        )

        close_range_subsweep = a121.SubsweepConfig(
            start_point=start_point,
            num_points=num_points,
            step_length=step_length,
            profile=a121.Profile.PROFILE_1,
            hwaas=hwaas,
            receiver_gain=5,
            phase_enhancement=True,
            iq_imbalance_compensation=True,
            prf=a121.PRF.PRF_15_6_MHz,
        )
        sensor_config = a121.SensorConfig(
            subsweeps=[loopback_subsweep, close_range_subsweep],
            sweeps_per_frame=10,
        )

        groups.append({sensor_id: sensor_config for sensor_id in sensor_ids})

    # TRANSITION
    far_subsweeps = []

    if range_types & RangeTypes.TRANSITION_P1:
        bounding_profile = min(a121.Profile.PROFILE_3, config.max_profile, key=lambda p: p.value)

        if range_types.closest_range_type is RangeTypes.TRANSITION_P1:
            start_m = config.start_m
        else:
            start_m = max(config.start_m, min_dist_m[a121.Profile.PROFILE_1])

        end_m = min(config.end_m, min_dist_m[bounding_profile])
        step_length = _limit_step_length(a121.Profile.PROFILE_1, config.max_step_length)

        (left_filter_margin, right_filter_margin) = _transition_p1_margins(
            step_length, range_types
        )

        if config.threshold_method is ThresholdMethod.CFAR:
            cfar_margin = _cfar_margins(a121.Profile.PROFILE_1, step_length)
        else:
            cfar_margin = 0

        marginless_start_point = _m_to_point(start_m, start_m, step_length)
        marginless_end_point = _m_to_point(start_m, end_m, step_length)

        if marginless_start_point == marginless_end_point:
            marginless_end_point += step_length

        start_point = marginless_start_point - left_filter_margin - cfar_margin
        end_point = marginless_end_point + right_filter_margin + cfar_margin
        num_points = int((end_point - start_point) / step_length)

        hwaas = _calc_hwaas(
            a121.Profile.PROFILE_1,
            step_length,
            marginless_end_point * APPROX_BASE_STEP_LENGTH_M,
            config.signal_quality,
            config.reflector_shape.exponent,
        )

        far_subsweeps.append(
            a121.SubsweepConfig(
                start_point=start_point,
                num_points=num_points,
                step_length=step_length,
                profile=a121.Profile.PROFILE_1,
                hwaas=hwaas,
                receiver_gain=10,
                phase_enhancement=True,
                iq_imbalance_compensation=True,
                prf=prf,
            )
        )

    if range_types & RangeTypes.TRANSITION_P3:
        start_m = max(config.start_m, min_dist_m[a121.Profile.PROFILE_3])
        end_m = min(config.end_m, min_dist_m[config.max_profile])
        step_length = _limit_step_length(a121.Profile.PROFILE_3, config.max_step_length)

        (left_filter_margin, right_filter_margin) = _transition_p3_margins(
            step_length,
            range_types,
        )
        if config.threshold_method is ThresholdMethod.CFAR:
            cfar_margin = _cfar_margins(a121.Profile.PROFILE_3, step_length)
        else:
            cfar_margin = 0

        marginless_start_point = _m_to_point(start_m, start_m, step_length)
        marginless_end_point = _m_to_point(start_m, end_m, step_length)

        if marginless_start_point == marginless_end_point:
            marginless_end_point += step_length

        start_point = marginless_start_point - left_filter_margin - cfar_margin
        end_point = marginless_end_point + right_filter_margin + cfar_margin
        num_points = int((end_point - start_point) / step_length)

        hwaas = _calc_hwaas(
            a121.Profile.PROFILE_3,
            step_length,
            marginless_end_point * APPROX_BASE_STEP_LENGTH_M,
            config.signal_quality,
            config.reflector_shape.exponent,
        )

        far_subsweeps.append(
            a121.SubsweepConfig(
                start_point=start_point,
                num_points=num_points,
                step_length=step_length,
                profile=a121.Profile.PROFILE_3,
                hwaas=hwaas,
                receiver_gain=10,
                phase_enhancement=True,
                iq_imbalance_compensation=True,
                prf=prf,
            )
        )

    # Max Profile
    if range_types & RangeTypes.MAX_PROFILE:
        if range_types.closest_range_type is RangeTypes.MAX_PROFILE:
            start_m = config.start_m
        else:
            start_m = max(config.start_m, min_dist_m[config.max_profile])

        end_m = config.end_m
        step_length = _limit_step_length(config.max_profile, config.max_step_length)

        (left_filter_margin, right_filter_margin) = _max_profile_margins(
            step_length, config.max_profile, range_types
        )
        if config.threshold_method is ThresholdMethod.CFAR:
            cfar_margin = _cfar_margins(config.max_profile, step_length)
        else:
            cfar_margin = 0

        num_max_profile_subsweeps = _NUM_SUBSWEEPS_IN_SENSOR_CONFIG - len(
            range_types & RangeTypes.ANY_TRANSITION
        )

        breakpoints_m = np.linspace(
            start_m,
            end_m,
            num_max_profile_subsweeps + 1,
        ).tolist()

        breakpoints = [_m_to_point(start_m, bp, step_length) for bp in breakpoints_m]

        breakpoints = remove_dup(breakpoints)

        num_max_profile_subsweeps = len(breakpoints) - 1

        for i in range(num_max_profile_subsweeps):
            marginless_start_point = breakpoints[i]
            marginless_end_point = breakpoints[i + 1]
            start_point = marginless_start_point
            end_point = marginless_end_point

            if i == 0:
                start_point = marginless_start_point - (left_filter_margin + cfar_margin)
            if i == (num_max_profile_subsweeps - 1):
                end_point = marginless_end_point + (right_filter_margin + cfar_margin)

            num_points = int((end_point - start_point) / step_length)

            hwaas = _calc_hwaas(
                config.max_profile,
                step_length,
                marginless_end_point * APPROX_BASE_STEP_LENGTH_M,
                config.signal_quality,
                config.reflector_shape.exponent,
            )

            far_subsweeps.append(
                a121.SubsweepConfig(
                    start_point=start_point,
                    num_points=num_points,
                    step_length=step_length,
                    profile=config.max_profile,
                    hwaas=hwaas,
                    receiver_gain=10,
                    phase_enhancement=True,
                    iq_imbalance_compensation=True,
                    prf=prf,
                )
            )

    if len(far_subsweeps) > 0:
        sensor_config = a121.SensorConfig(subsweeps=far_subsweeps, sweeps_per_frame=1)
        groups.append({sensor_id: sensor_config for sensor_id in sensor_ids})

    return a121.SessionConfig(groups, extended=True, update_rate=config.update_rate)


def get_num_far_subsweeps(
    session_config: a121.SessionConfig,
    config: DetectorConfig,
) -> int:
    range_types = _get_range_types(config)
    if RangeTypes.CLOSE_RANGE & range_types and len(session_config.groups) == 1:
        # no far group
        return 0

    far_sensor_group = session_config.groups[-1]
    # same config for each sensor. Grab first sensor id.
    num_far_subsweeps = far_sensor_group[list(far_sensor_group.keys())[0]].num_subsweeps
    return num_far_subsweeps


def _get_furthest_measurement(config: DetectorConfig) -> tuple[float, a121.Profile]:
    range_types = _get_range_types(config)

    if RangeTypes.MAX_PROFILE in range_types:
        step_length = _limit_step_length(config.max_profile, config.max_step_length)

        (_, right_filter_margin) = _max_profile_margins(
            step_length, config.max_profile, range_types
        )

        if config.threshold_method is ThresholdMethod.CFAR:
            cfar_margin = _cfar_margins(config.max_profile, step_length)
        else:
            cfar_margin = 0

        distance_m = config.end_m + (right_filter_margin + cfar_margin) * APPROX_BASE_STEP_LENGTH_M
        return (
            distance_m,
            config.max_profile,
        )

    if RangeTypes.TRANSITION_P3 in range_types:
        step_length = _limit_step_length(a121.Profile.PROFILE_3, config.max_step_length)

        (_, right_filter_margin) = _transition_p3_margins(step_length, range_types)

        if config.threshold_method is ThresholdMethod.CFAR:
            cfar_margin = _cfar_margins(a121.Profile.PROFILE_3, step_length)
        else:
            cfar_margin = 0

        distance_m = config.end_m + (right_filter_margin + cfar_margin) * APPROX_BASE_STEP_LENGTH_M
        return (
            distance_m,
            a121.Profile.PROFILE_3,
        )

    if RangeTypes.TRANSITION_P1 in range_types:
        step_length = _limit_step_length(a121.Profile.PROFILE_1, config.max_step_length)

        (_, right_filter_margin) = _transition_p1_margins(step_length, range_types)

        if config.threshold_method is ThresholdMethod.CFAR:
            cfar_margin = _cfar_margins(a121.Profile.PROFILE_1, step_length)
        else:
            cfar_margin = 0

        distance_m = config.end_m + (right_filter_margin + cfar_margin) * APPROX_BASE_STEP_LENGTH_M
        return (
            distance_m,
            a121.Profile.PROFILE_1,
        )

    if RangeTypes.CLOSE_RANGE in range_types:
        step_length = _limit_step_length(a121.Profile.PROFILE_1, config.max_step_length)
        (_, right_filter_margin) = _close_range_margins(step_length, range_types)

        distance_m = config.end_m + (right_filter_margin * APPROX_BASE_STEP_LENGTH_M)
        return (
            distance_m,
            a121.Profile.PROFILE_1,
        )

    msg = "Unreachable"
    raise Exception(msg)


def _get_max_prf(config: DetectorConfig) -> a121.PRF:
    (distance_m, profile) = _get_furthest_measurement(config)
    prf = select_prf_m(distance_m, profile)
    return prf


def _m_to_point(start_m: float, distance_m: float, step_length: int) -> int:
    start_p = int(round(start_m / APPROX_BASE_STEP_LENGTH_M))
    point_frac = (distance_m - start_m) / APPROX_BASE_STEP_LENGTH_M + start_p
    return (round(point_frac) // step_length) * step_length


def remove_dup(breakpoints: t.List[int]) -> t.List[int]:
    unique_bps = sorted(set(breakpoints[1:]))

    unique_bps = [breakpoints[0]] + unique_bps

    return unique_bps


def _calc_leakage_free_min_dist(config: DetectorConfig) -> dict[a121.Profile, float]:
    """This function calculates the shortest leakage free distance per profile, for all profiles
    up to max_profile"""
    min_dist_m = {}
    for profile, min_dist in _MIN_LEAKAGE_FREE_DIST_M.items():
        min_dist_m[profile] = min_dist
        if config.threshold_method == ThresholdMethod.CFAR:
            step_length = _limit_step_length(profile, config.max_step_length)
            cfar_margin_m = (
                Processor.calc_cfar_margin(profile, step_length)
                * step_length
                * APPROX_BASE_STEP_LENGTH_M
            )
            min_dist_m[profile] += cfar_margin_m

        if profile == config.max_profile:
            # All profiles up to max_profile has been added. Break and return result.
            break

    return min_dist_m


def _limit_step_length(profile: a121.Profile, user_limit: t.Optional[int]) -> int:
    """
    Calculates step length based on user defined step length and selected profile.

    The step length must yield minimum _MIN_NUM_POINTS_IN_ENVELOPE_FWHM_SPAN number of points
    in the span of the FWHM of the envelope.

    If the step length is <24, return the valid step length(defined by
    _VALID_STEP_LENGTHS) that is closest to, but not longer than the limit.

    If the limit is 24<=, return the multiple of 24 that is
    closest, but not longer than the limit.
    """

    fwhm_p = ENVELOPE_FWHM_M[profile] / APPROX_BASE_STEP_LENGTH_M
    limit = int(fwhm_p / _MIN_NUM_POINTS_IN_ENVELOPE_FWHM_SPAN)

    if user_limit is not None:
        limit = min(user_limit, limit)

    if limit < _VALID_STEP_LENGTHS[-1]:
        idx_closest = np.sum(np.array(_VALID_STEP_LENGTHS) <= limit) - 1
        return int(_VALID_STEP_LENGTHS[idx_closest])
    else:
        return int((limit // _VALID_STEP_LENGTHS[-1]) * _VALID_STEP_LENGTHS[-1])



# Copyright (c) Acconeer AB, 2023
# All rights reserved

from ._example_app import DetectionState, ExampleApp, ExampleAppConfig, ExampleAppResult
from ._mode_handler import AppMode, ModeHandler, ModeHandlerConfig, ModeHandlerResult



# Copyright (c) Acconeer AB, 2025
# All rights reserved

from __future__ import annotations

from typing import Any, Optional

from acconeer.exptool.a121._core.entities import SensorCalibration, SessionConfig
from acconeer.exptool.a121._core.utils import (
    iterate_extended_structure_values,
    map_over_extended_structure,
)

from ._latest import ExplorationProtocol, ExplorationProtocolError


class ExplorationProtocol_No_IQ_Imb_Comp(ExplorationProtocol):
    @classmethod
    def setup_command(
        cls,
        session_config: SessionConfig,
        calibrations: Optional[dict[int, SensorCalibration]] = None,
    ) -> bytes:
        for sensor_config in iterate_extended_structure_values(session_config.groups):
            if any(ssc.iq_imbalance_compensation for ssc in sensor_config.subsweeps):
                msg = "Connected Exploration server does not support IQ imbalance compensation."
                raise ExplorationProtocolError(msg)

        return super().setup_command(session_config, calibrations)

    @classmethod
    def _setup_command_preprocessing(cls, session_config: SessionConfig) -> dict[str, Any]:
        result = super()._setup_command_preprocessing(session_config)
        result["groups"] = map_over_extended_structure(cls._remove_iq_imb_comp, result["groups"])

        return result

    @classmethod
    def _remove_iq_imb_comp(cls, sensor_config_dict: dict[str, Any]) -> dict[str, Any]:
        for subsweep_config_dict in sensor_config_dict["subsweeps"]:
            subsweep_config_dict.pop("iq_imbalance_compensation")
        return sensor_config_dict



# Copyright (c) Acconeer AB, 2024
# All rights reserved

from ._configs import (
    get_processor_config,
    get_sensor_config,
)
from ._processor import (
    Processor,
    ProcessorConfig,
    ProcessorContext,
    ProcessorResult,
)



# Copyright (c) Acconeer AB, 2022-2025
# All rights reserved

from __future__ import annotations

import logging
from enum import Enum
from functools import partial
from importlib.resources import as_file, files
from typing import Any, Optional

import qtawesome as qta

from PySide6 import QtCore, QtGui
from PySide6.QtSvgWidgets import QSvgWidget
from PySide6.QtWidgets import (
    QButtonGroup,
    QFrame,
    QGraphicsOpacityEffect,
    QGridLayout,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QSizePolicy,
    QVBoxLayout,
    QWidget,
)

from acconeer.exptool.app import resources
from acconeer.exptool.app.new._enums import PluginFamily, PluginGeneration, PluginState
from acconeer.exptool.app.new.app_model import AppModel, PluginPresetSpec, PluginSpec
from acconeer.exptool.app.new.pluginbase import PlotPluginBase, PluginSpecBase
from acconeer.exptool.app.new.ui.components.group_box import GroupBox
from acconeer.exptool.app.new.ui.icons import ARROW_LEFT_BOLD, EXTERNAL_LINK, TEXT_GREY
from acconeer.exptool.app.new.ui.utils import (
    HorizontalSeparator,
    LayoutWrapper,
    ScrollAreaDecorator,
    TopAlignDecorator,
)


log = logging.getLogger(__name__)


class PluginSelectionButton(QPushButton):
    plugin: PluginSpec

    def __init__(self, plugin: PluginSpecBase, parent: QWidget) -> None:
        super().__init__(parent)

        self.plugin = plugin

        self.setText(plugin.title)
        self.setStyleSheet("text-align: left; font-weight: bold;")
        self.setCheckable(True)


class PluginSelectionButtonGroup(QButtonGroup):
    def __init__(self, parent: QWidget) -> None:
        super().__init__(parent)

        self.setExclusive(True)

    def checkedButton(self) -> PluginSelectionButton:
        button = super().checkedButton()
        assert isinstance(button, PluginSelectionButton)
        return button

    def buttons(self) -> list[PluginSelectionButton]:  # type: ignore[override]
        buttons = super().buttons()
        assert isinstance(buttons, list)
        assert all(isinstance(e, PluginSelectionButton) for e in buttons)
        return buttons  # type: ignore[return-value]


class PluginPresetButton(QPushButton):
    def __init__(self, plugin_preset: PluginPresetSpec, default: bool, parent: QWidget) -> None:
        super().__init__(parent)

        if default:
            self.setText(f"{plugin_preset.name} (default)")
        else:
            self.setText(plugin_preset.name)
        if plugin_preset.description is not None:
            self.setToolTip(plugin_preset.description)
        self.setStyleSheet("text-align: left; font-weight: bold;")


class PluginPresetPlaceholder(QWidget):
    def __init__(self, app_model: AppModel, parent: Optional[QWidget] = None) -> None:
        super().__init__(parent)

        self.app_model = app_model
        self.app_model.sig_load_plugin.connect(self._on_load_plugin)
        self.app_model.sig_notify.connect(self._on_app_model_update)

        self.app_model = app_model
        layout = QVBoxLayout(self)

        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(11)
        self.group_box = GroupBox.vertical("Preset Configurations", parent=self)
        layout.addWidget(self.group_box)
        self.preset_buttons_widget: Optional[QWidget] = None

        self.setLayout(layout)

    def _clear(self) -> None:
        if self.preset_buttons_widget is not None:
            self.preset_buttons_widget.deleteLater()
            self.preset_buttons_widget = None

    def _on_preset_click(self, preset_id: Enum) -> None:
        self.app_model.set_plugin_preset(preset_id)

    def _on_load_plugin(self, plugin_spec: Optional[PluginSpecBase]) -> None:
        self._clear()
        if plugin_spec is not None and plugin_spec.presets is not None:
            self.preset_buttons_widget = QWidget(self.group_box)
            preset_buttons_layout = QVBoxLayout(self.preset_buttons_widget)
            preset_buttons_layout.setContentsMargins(0, 0, 0, 0)

            self.group_box.layout().addWidget(self.preset_buttons_widget)

            for preset in plugin_spec.presets:
                default_preset = (
                    len(plugin_spec.presets) > 1
                    and preset.preset_id == plugin_spec.default_preset_id
                    and preset.name.lower() != "default"
                )
                button = PluginPresetButton(preset, default_preset, self)
                preset_buttons_layout.addWidget(button)
                button.clicked.connect(partial(self._on_preset_click, preset.preset_id))
            self.preset_buttons_widget.setLayout(preset_buttons_layout)

    def _on_app_model_update(self, app_model: AppModel) -> None:
        self.setEnabled(app_model.plugin_state == PluginState.LOADED_IDLE)


class PluginSelection(QWidget):
    PLUGIN_FAMILY_ORDER = [
        PluginFamily.EXTERNAL_PLUGIN,
        PluginFamily.SERVICE,
        PluginFamily.DETECTOR,
        PluginFamily.REF_APP,
        PluginFamily.EXAMPLE_APP,
    ]

    def __init__(
        self, app_model: AppModel, plugins: list[PluginSpecBase], parent: QWidget
    ) -> None:
        super().__init__(parent)

        self.app_model = app_model

        app_model.sig_notify.connect(self._on_app_model_update)
        layout = QVBoxLayout(self)

        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(11)

        group_boxes = {}
        for family in self.PLUGIN_FAMILY_ORDER:
            group_box = QGroupBox(self)
            group_box.setTitle(family.value)
            group_box.setHidden(True)
            group_box.setLayout(QVBoxLayout(group_box))
            layout.addWidget(group_box)
            group_boxes[family] = group_box

        self.button_group = PluginSelectionButtonGroup(self)
        self.button_group.buttonClicked.connect(self._on_load_click)

        for plugin in plugins:
            assert isinstance(plugin, PluginSpecBase)
            group_box = group_boxes[plugin.family]
            group_box.setHidden(False)

            button_layout = QHBoxLayout()
            button_layout.setContentsMargins(0, 0, 0, 0)

            group_box_layout = group_box.layout()
            assert group_box_layout is not None
            group_box_layout.addWidget(LayoutWrapper(button_layout))

            button = PluginSelectionButton(plugin, group_box)
            self.button_group.addButton(button)
            button_layout.addWidget(button)

            if plugin.description:
                label = QLabel(group_box)
                label.setText(plugin.description)
                label.setWordWrap(True)
                group_box_layout.addWidget(label)

            if plugin.docs_link:
                info_button = QPushButton(QtGui.QIcon(EXTERNAL_LINK(color=TEXT_GREY)), "docs")
                info_button.setStyleSheet(f"QPushButton {{color: {TEXT_GREY}}}")
                info_button.setSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
                info_button.clicked.connect(
                    lambda _=False, p=plugin: QtGui.QDesktopServices.openUrl(p.docs_link)
                )
                button_layout.addWidget(info_button)
        self.setLayout(layout)

    def layout(self) -> QVBoxLayout:
        return super().layout()  # type: ignore[return-value]

    def _on_load_click(self) -> None:
        self.layout().addStretch(1)

        plugin = self.button_group.checkedButton().plugin
        self.app_model.load_plugin(plugin)

    def _on_app_model_update(self, app_model: AppModel) -> None:
        plugin: Optional[PluginSpec] = app_model.plugin

        if plugin is None:
            self.button_group.setExclusive(False)

            for button in self.button_group.buttons():
                button.setChecked(False)

            self.button_group.setExclusive(True)
        else:
            buttons = self.button_group.buttons()
            try:
                (button_to_check,) = {b for b in buttons if b.plugin == plugin}
            except ValueError:
                pass
            else:
                button_to_check.setChecked(True)

        self.setEnabled(app_model.plugin_state.is_steady)


class PluginSelectionArea(QWidget):
    def __init__(self, app_model: AppModel, parent: Optional[QWidget] = None) -> None:
        super().__init__(parent)

        layout = QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)

        a121_selection = ScrollAreaDecorator(
            TopAlignDecorator(
                PluginSelection(
                    app_model,
                    [
                        plugin
                        for plugin in app_model.plugins
                        if (
                            plugin.generation == PluginGeneration.A121
                            and isinstance(plugin, PluginSpecBase)
                        )
                    ],
                    self,
                )
            )
        )
        a121_selection.setVisible(True)

        app_model.sig_notify.connect(
            lambda am: a121_selection.setVisible(am.plugin_generation == PluginGeneration.A121)
        )

        layout.addWidget(a121_selection)
        layout.addWidget(HorizontalSeparator())
        layout.addWidget(PluginPresetPlaceholder(app_model, self))
        self.setLayout(layout)


class PlotPlaceholder(PlotPluginBase):
    def __init__(self, app_model: AppModel) -> None:
        super().__init__(app_model)

        layout = QVBoxLayout(self)

        layout.addStretch(1)
        layout.addLayout(self._select_module_text())
        layout.addWidget(self._teaser_text())
        layout.addStretch(1)

        self.setLayout(layout)

    def _select_module_text(self) -> QHBoxLayout:
        h_box = QHBoxLayout()
        h_box.addStretch(1)

        icon_widget = qta.IconWidget()
        icon_widget.setIconSize(QtCore.QSize(36, 36))
        icon_widget.setIcon(ARROW_LEFT_BOLD(color="#4d5157"))

        label = QLabel("Select a module to begin", self)
        label.setStyleSheet("font-size: 20px;")
        label.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)

        h_box.addWidget(icon_widget)
        h_box.addWidget(label)
        h_box.addStretch(1)

        return h_box

    def _teaser_text(self) -> QLabel:
        teaser_label = QLabel(
            "More detectors and example applications will\nbe released continuously.", self
        )
        teaser_label.setStyleSheet("font-size: 16px;")
        teaser_label.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)

        return teaser_label

    def handle_message(self, message: Any) -> None:
        pass

    def draw(self) -> None:
        pass


class PluginPlotArea(QFrame):
    _FPS = 60

    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        self.app_model = app_model

        self.plot_plugin: PlotPluginBase = PlotPlaceholder(app_model)

        self.setObjectName("PluginPlotArea")
        self.setStyleSheet("QFrame#PluginPlotArea {background: #fff; border: 0;}")
        self.setFrameStyle(0)

        layout = QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        layout.addWidget(self.plot_plugin)

        self.startTimer(int(1000 / self._FPS))

        app_model.sig_load_plugin.connect(self._on_app_model_load_plugin)
        if isinstance(app_model.plugin, PluginSpecBase):
            self._on_app_model_load_plugin(app_model.plugin)
        elif app_model.plugin is not None:
            msg = f"{type(app_model.plugin)} is not a PluginSpecBase."
            raise RuntimeError(msg)

        self.setLayout(layout)

    def timerEvent(self, event: QtCore.QTimerEvent) -> None:
        plugin_class_name = type(self.plot_plugin).__name__
        with self.app_model.report_timing(f"{plugin_class_name}.draw()"):
            self.plot_plugin.draw()

    def _on_app_model_load_plugin(self, plugin: Optional[PluginSpecBase]) -> None:
        log.debug(
            f"{self.__class__.__name__} is going to replace its plot_plugin "
            + f"({self.plot_plugin.__class__.__name__}) in favour of {plugin}"
        )
        self.plot_plugin.deleteLater()

        if plugin is None:
            self.plot_plugin = PlotPlaceholder(self.app_model)
        else:
            self.plot_plugin = plugin.create_plot_plugin(app_model=self.app_model)

        layout = self.layout()
        assert layout is not None
        layout.addWidget(self.plot_plugin)


class ControlPlaceholder(QWidget):
    def __init__(self, parent: Optional[QWidget] = None) -> None:
        super().__init__(parent)

        layout = QGridLayout(self)

        with as_file(files(resources) / "icon-black.svg") as path:
            icon = QSvgWidget(str(path), self)

        icon.setMaximumSize(60, 60)
        icon.renderer().setAspectRatioMode(QtCore.Qt.AspectRatioMode.KeepAspectRatio)
        effect = QGraphicsOpacityEffect(icon)
        effect.setOpacity(0.1)
        icon.setGraphicsEffect(effect)

        layout.addWidget(icon)
        self.setLayout(layout)


class PluginControlArea(QWidget):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        self.app_model = app_model
        self.app_model.sig_load_plugin.connect(self._on_app_model_load_plugin)

        self.child_widget: QWidget = ControlPlaceholder()
        self.view_plugin: Optional[Any] = None

        layout = QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        layout.addWidget(self.child_widget)

        if isinstance(app_model.plugin, PluginSpecBase):
            self._on_app_model_load_plugin(app_model.plugin)
        elif app_model.plugin is not None:
            msg = f"{type(app_model.plugin)} is not a PluginSpecBase."
            raise RuntimeError(msg)
        self.setLayout(layout)

    def _on_app_model_load_plugin(self, plugin_spec: Optional[PluginSpecBase]) -> None:
        log.debug(
            f"{self.__class__.__name__} is going to replace its view_plugin"
            + f"({self.view_plugin.__class__.__name__}) in favour of {plugin_spec}"
        )
        self.view_plugin = None
        self.child_widget.deleteLater()

        if plugin_spec is None:
            self.child_widget = ControlPlaceholder()
        else:
            self.child_widget = plugin_spec.create_view_plugin(
                app_model=self.app_model,
            )

        layout = self.layout()
        assert layout is not None
        layout.addWidget(self.child_widget)



# Copyright (c) Acconeer AB, 2023-2026
# All rights reserved

from __future__ import annotations

from acconeer.exptool import a121
from acconeer.exptool._core.docstrings import get_attribute_docstring
from acconeer.exptool.app.new.ui.components import PidgetFactoryMapping, pidgets
from acconeer.exptool.app.new.ui.components.pidgets.hooks import (
    disable_if,
    enable_if,
    parameter_is,
)

from ._detector import DetectorConfig, PeakSortingMethod, ReflectorShape, ThresholdMethod


def get_pidget_mapping() -> PidgetFactoryMapping:
    return {
        "start_m": pidgets.FloatPidgetFactory(
            name_label_text="Range start:",
            name_label_tooltip=get_attribute_docstring(DetectorConfig, "start_m"),
            suffix=" m",
            decimals=3,
        ),
        "end_m": pidgets.FloatPidgetFactory(
            name_label_text="Range end:",
            name_label_tooltip=get_attribute_docstring(DetectorConfig, "end_m"),
            suffix=" m",
            decimals=3,
        ),
        "close_range_leakage_cancellation": pidgets.CheckboxPidgetFactory(
            name_label_text="Enable close range leakage cancellation",
            name_label_tooltip=get_attribute_docstring(
                DetectorConfig, "close_range_leakage_cancellation"
            ),
        ),
        "max_step_length": pidgets.OptionalIntPidgetFactory(
            name_label_text="Max step length:",
            name_label_tooltip=get_attribute_docstring(DetectorConfig, "max_step_length"),
            checkbox_label_text="Set",
            limits=(1, None),
            init_set_value=12,
        ),
        "max_profile": pidgets.EnumPidgetFactory(
            name_label_text="Max profile:",
            name_label_tooltip=get_attribute_docstring(DetectorConfig, "max_profile"),
            enum_type=a121.Profile,
            label_mapping={
                a121.Profile.PROFILE_1: "1 (shortest)",
                a121.Profile.PROFILE_2: "2",
                a121.Profile.PROFILE_3: "3",
                a121.Profile.PROFILE_4: "4",
                a121.Profile.PROFILE_5: "5 (longest)",
            },
        ),
        "reflector_shape": pidgets.EnumPidgetFactory(
            name_label_text="Reflector shape:",
            name_label_tooltip=get_attribute_docstring(DetectorConfig, "reflector_shape"),
            enum_type=ReflectorShape,
            label_mapping={
                ReflectorShape.GENERIC: "Generic",
                ReflectorShape.PLANAR: "Planar",
            },
        ),
        "peaksorting_method": pidgets.EnumPidgetFactory(
            name_label_text="Peak sorting method:",
            name_label_tooltip=get_attribute_docstring(DetectorConfig, "peaksorting_method"),
            enum_type=PeakSortingMethod,
            label_mapping={
                PeakSortingMethod.CLOSEST: "Closest",
                PeakSortingMethod.STRONGEST: "Strongest",
            },
        ),
        "threshold_method": pidgets.EnumPidgetFactory(
            name_label_text="Threshold method:",
            name_label_tooltip=get_attribute_docstring(DetectorConfig, "threshold_method"),
            enum_type=ThresholdMethod,
            label_mapping={
                ThresholdMethod.CFAR: "CFAR",
                ThresholdMethod.FIXED: "Fixed amplitude",
                ThresholdMethod.FIXED_STRENGTH: "Fixed strength",
                ThresholdMethod.RECORDED: "Recorded",
            },
        ),
        "fixed_threshold_value": pidgets.FloatPidgetFactory(
            name_label_text="Fixed amplitude threshold value:",
            name_label_tooltip=get_attribute_docstring(DetectorConfig, "fixed_threshold_value"),
            decimals=1,
            limits=(0, None),
            hooks=enable_if(parameter_is("threshold_method", ThresholdMethod.FIXED)),
        ),
        "fixed_strength_threshold_value": pidgets.FloatPidgetFactory(
            name_label_text="Fixed strength threshold value:",
            name_label_tooltip=get_attribute_docstring(
                DetectorConfig, "fixed_strength_threshold_value"
            ),
            decimals=1,
            hooks=enable_if(parameter_is("threshold_method", ThresholdMethod.FIXED_STRENGTH)),
            suffix="dBsm",
        ),
        "num_frames_in_recorded_threshold": pidgets.IntPidgetFactory(
            name_label_text="Num frames in rec. thr.:",
            name_label_tooltip=get_attribute_docstring(
                DetectorConfig, "num_frames_in_recorded_threshold"
            ),
            limits=(1, None),
        ),
        "threshold_sensitivity": pidgets.FloatSliderPidgetFactory(
            name_label_text="Threshold sensitivity:",
            name_label_tooltip=get_attribute_docstring(DetectorConfig, "threshold_sensitivity"),
            decimals=2,
            limits=(0, 1),
            show_limit_values=False,
            hooks=disable_if(
                parameter_is("threshold_method", ThresholdMethod.FIXED),
                parameter_is("threshold_method", ThresholdMethod.FIXED_STRENGTH),
            ),
        ),
        "signal_quality": pidgets.FloatSliderPidgetFactory(
            name_label_text="Signal quality:",
            name_label_tooltip=get_attribute_docstring(DetectorConfig, "signal_quality"),
            decimals=1,
            limits=(-10, 35),
            show_limit_values=False,
            limit_texts=("Less power", "Higher quality"),
        ),
        "update_rate": pidgets.OptionalFloatPidgetFactory(
            name_label_tooltip=get_attribute_docstring(DetectorConfig, "update_rate"),
            name_label_text="Update rate:",
            checkbox_label_text="Set",
            limits=(1, None),
            init_set_value=20,
        ),
    }



# Copyright (c) Acconeer AB, 2024
# All rights reserved

from __future__ import annotations

import copy
from typing import Any, Dict, List, Optional, Tuple

import attrs
import h5py
import numpy as np
import numpy.typing as npt
import typing_extensions as te

from acconeer.exptool import a121, opser
from acconeer.exptool import type_migration as tm
from acconeer.exptool._core.class_creation.attrs import attrs_optional_ndarray_isclose
from acconeer.exptool.a121._h5_utils import _create_h5_string_dataset
from acconeer.exptool.a121.algo import (
    AlgoBase,
    calculate_loopback_peak_location,
)

from ._aggregator import Aggregator, AggregatorConfig, ProcessorSpec
from ._processors import ProcessorContext, ProcessorMode, calculate_bg_noise_std
from ._utils import (
    filter_close_range_spec,
    get_calibrate_noise_session_config,
    get_calibrate_offset_sensor_config,
    update_processor_mode,
)


@attrs.mutable(kw_only=True)
class DetectorContext(AlgoBase):
    offset_calibration: Optional[OffsetCalibration] = attrs.field(default=None)
    noise_calibration: Optional[NoiseCalibration] = attrs.field(default=None)
    close_range_calibration: Optional[CloseRangeCalibration] = attrs.field(default=None)
    recorded_threshold_calibration: Optional[RecordedThresholdCalibration] = attrs.field(
        default=None
    )
    sensor_ids: List[int] = attrs.field(factory=list)
    session_config_used_during_calibration: Optional[a121.SessionConfig] = attrs.field(
        default=None
    )


@attrs.frozen(kw_only=True)
class OffsetCalibration:
    """Estimates sensor offset error based on loopback measurement."""

    results: List[Dict[int, a121.Result]]

    @classmethod
    def create(cls, client: a121.Client, sensor_ids: List[int]) -> te.Self:
        sensor_config = get_calibrate_offset_sensor_config()

        session_config = a121.SessionConfig(
            {sensor_id: sensor_config for sensor_id in sensor_ids}, extended=True
        )
        client.setup_session(session_config)
        client.start_session()
        extended_result = client.get_next()
        client.stop_session()

        assert isinstance(extended_result, list)

        return cls(results=extended_result)

    def loopback_peak_location_m(self, sensor_id: int) -> float:
        sensor_config = get_calibrate_offset_sensor_config()
        lb_peak_location_m = calculate_loopback_peak_location(
            self.results[0][sensor_id], sensor_config
        )
        return lb_peak_location_m


@attrs.frozen(kw_only=True)
class NoiseCalibration:
    """Estimates the standard deviation of the noise in each subsweep by setting enable_tx to
    False and collecting data, used to calculate the deviation.

    The calibration procedure can be done at any time as it is performed with Tx off, and is
    not effected by objects in front of the sensor.

    This function is called from the start() in the case of CFAR and Fixed threshold and from
    record_threshold() in the case of Recorded threshold. The reason for calling from
    record_threshold() is that it is used when calculating the threshold.
    """

    results: List[Dict[int, a121.Result]]

    @classmethod
    def create(
        cls, client: a121.Client, sensor_ids: List[int], session_config: a121.SessionConfig
    ) -> te.Self:
        session_config = get_calibrate_noise_session_config(session_config, sensor_ids)

        extended_metadata = client.setup_session(session_config)
        assert isinstance(extended_metadata, list)

        client.start_session()
        extended_result = client.get_next()
        client.stop_session()

        assert isinstance(extended_result, list)
        return cls(results=extended_result)

    def bg_noise_std(
        self,
        sensor_id: int,
        processor_specs: List[ProcessorSpec],
        session_config: a121.SessionConfig,
    ) -> List[List[float]]:
        bg_noise_one_sensor = []
        for spec in processor_specs:
            result = self.results[spec.group_index][sensor_id]
            sensor_config = session_config.groups[spec.group_index][sensor_id]
            subsweep_configs = sensor_config.subsweeps
            bg_noise_std_in_subsweep = []
            for idx in spec.subsweep_indexes:
                if not subsweep_configs[idx].enable_loopback:
                    subframe = result.subframes[idx]
                    subsweep_std = calculate_bg_noise_std(subframe, subsweep_configs[idx])
                    bg_noise_std_in_subsweep.append(subsweep_std)
            bg_noise_one_sensor.append(bg_noise_std_in_subsweep)

        return bg_noise_one_sensor


@attrs.frozen(kw_only=True)
class CloseRangeCalibration:
    """Calibrates the close range measurement parameters used when subtracting the direct
    leakage from the measured signal.

    The parameters calibrated are the direct leakage and a phase reference, used to reduce
    the amount of phase jitter, with the purpose of reducing the residual.
    """

    extended_metadata: List[Dict[int, a121.Metadata]]
    results: List[Dict[int, a121.Result]]
    sensor_calibrations: Dict[int, a121.SensorCalibration]

    @classmethod
    def create(cls, client: a121.Client, session_config: a121.SessionConfig) -> te.Self:
        extended_metadata = client.setup_session(session_config)
        assert isinstance(extended_metadata, list)

        client.start_session()
        extended_result = client.get_next()
        client.stop_session()

        assert isinstance(extended_result, list)

        return cls(
            extended_metadata=extended_metadata,
            results=extended_result,
            sensor_calibrations=client.calibrations,
        )

    def _close_range_calibration(
        self,
        sensor_id: int,
        processor_specs: List[ProcessorSpec],
        session_config: a121.SessionConfig,
    ) -> Tuple[npt.NDArray[np.complex128], npt.NDArray[np.float64]]:
        close_range_spec = filter_close_range_spec(processor_specs)
        spec = update_processor_mode(close_range_spec, ProcessorMode.LEAKAGE_CALIBRATION)

        aggregator = Aggregator(
            session_config=session_config,
            extended_metadata=self.extended_metadata,
            config=AggregatorConfig(),
            specs=spec,
            sensor_id=sensor_id,
        )

        aggregator_result = aggregator.process(extended_result=self.results)
        (processor_result,) = aggregator_result.processor_results
        assert processor_result.direct_leakage is not None
        assert processor_result.phase_jitter_comp_reference is not None

        return (processor_result.direct_leakage, processor_result.phase_jitter_comp_reference)

    def direct_leakage(
        self,
        sensor_id: int,
        processor_specs: List[ProcessorSpec],
        session_config: a121.SessionConfig,
    ) -> npt.NDArray[np.complex128]:
        direct_leakage, _ = self._close_range_calibration(
            sensor_id, processor_specs, session_config
        )
        return direct_leakage

    def phase_jitter_comp_reference(
        self,
        sensor_id: int,
        processor_specs: List[ProcessorSpec],
        session_config: a121.SessionConfig,
    ) -> npt.NDArray[np.float64]:
        _, phase_jitter_comp_reference = self._close_range_calibration(
            sensor_id, processor_specs, session_config
        )
        return phase_jitter_comp_reference


@attrs.frozen(kw_only=True)
class RecordedThresholdCalibration:
    """Calibrates the parameters used when forming the recorded threshold."""

    extended_metadata: List[Dict[int, a121.Metadata]]
    results: List[List[Dict[int, a121.Result]]]

    @classmethod
    def create(
        cls,
        client: a121.Client,
        session_config: a121.SessionConfig,
        num_frames_in_recorded_threshold: int,
    ) -> te.Self:
        extended_metadata = client.setup_session(session_config)
        assert isinstance(extended_metadata, list)

        recorded_threshold_result = []

        client.start_session()
        for _ in range(num_frames_in_recorded_threshold):
            extended_result = client.get_next()
            assert isinstance(extended_result, list)
            recorded_threshold_result.append(extended_result)
        client.stop_session()

        return cls(extended_metadata=extended_metadata, results=recorded_threshold_result)

    @staticmethod
    def _add_context_to_processor_spec(
        noise_calibration: NoiseCalibration,
        close_range_calibration: Optional[CloseRangeCalibration],
        sensor_id: int,
        processor_specs: list[ProcessorSpec],
        session_config: a121.SessionConfig,
    ) -> List[ProcessorSpec]:
        """
        Create and add processor context to processor specification.
        """

        bg_noise_stds = noise_calibration.bg_noise_std(sensor_id, processor_specs, session_config)
        updated_specs: List[ProcessorSpec] = []

        for idx, (spec, bg_noise_std) in enumerate(zip(processor_specs, bg_noise_stds)):
            processor_context = ProcessorContext(
                recorded_threshold_mean_sweep=None,
                recorded_threshold_noise_std=None,
                bg_noise_std=bg_noise_std,
                direct_leakage=close_range_calibration.direct_leakage(
                    sensor_id,
                    processor_specs,
                    session_config,
                )
                if close_range_calibration is not None
                else None,
                phase_jitter_comp_ref=close_range_calibration.phase_jitter_comp_reference(
                    sensor_id,
                    processor_specs,
                    session_config,
                )
                if close_range_calibration is not None
                else None,
                reference_temperature=None,
                loopback_peak_location_m=None,
            )
            updated_specs.append(attrs.evolve(spec, processor_context=processor_context))
        return updated_specs

    def _recorded_threshold_calibration(
        self,
        sensor_id: int,
        processor_specs: List[ProcessorSpec],
        session_config: a121.SessionConfig,
        noise_calibration: NoiseCalibration,
        close_range_calibration: Optional[CloseRangeCalibration],
    ) -> Tuple[
        List[npt.NDArray[np.float64]],
        List[List[np.float64]],
        int,
    ]:
        specs_updated = update_processor_mode(
            processor_specs, ProcessorMode.RECORDED_THRESHOLD_CALIBRATION
        )

        specs = self._add_context_to_processor_spec(
            noise_calibration,
            close_range_calibration,
            sensor_id,
            specs_updated,
            session_config,
        )

        aggregator = Aggregator(
            session_config=session_config,
            extended_metadata=self.extended_metadata,
            config=AggregatorConfig(),
            specs=specs,
            sensor_id=sensor_id,
        )

        for extended_result in self.results:
            aggregator_result = aggregator.process(extended_result=extended_result)

        recorded_thresholds_mean_sweep = []
        recorded_thresholds_noise_std = []
        for processor_result in aggregator_result.processor_results:
            # Since we know what mode the processor is running in
            assert processor_result.recorded_threshold_mean_sweep is not None
            assert processor_result.recorded_threshold_noise_std is not None

            recorded_thresholds_mean_sweep.append(processor_result.recorded_threshold_mean_sweep)
            recorded_thresholds_noise_std.append(processor_result.recorded_threshold_noise_std)

        return (
            recorded_thresholds_mean_sweep,
            recorded_thresholds_noise_std,
            extended_result[0][sensor_id].temperature,
        )

    def recorded_thresholds_mean_sweep(
        self,
        sensor_id: int,
        processor_specs: List[ProcessorSpec],
        session_config: a121.SessionConfig,
        noise_calibration: NoiseCalibration,
        close_range_calibration: Optional[CloseRangeCalibration],
    ) -> List[npt.NDArray[np.float64]]:
        mean_sweep, _, _ = self._recorded_threshold_calibration(
            sensor_id,
            processor_specs,
            session_config,
            noise_calibration,
            close_range_calibration,
        )
        return mean_sweep

    def recorded_thresholds_noise_std(
        self,
        sensor_id: int,
        processor_specs: List[ProcessorSpec],
        session_config: a121.SessionConfig,
        noise_calibration: NoiseCalibration,
        close_range_calibration: Optional[CloseRangeCalibration],
    ) -> List[List[np.float64]]:
        _, noise_std, _ = self._recorded_threshold_calibration(
            sensor_id,
            processor_specs,
            session_config,
            noise_calibration,
            close_range_calibration,
        )
        return noise_std

    def reference_temperature(
        self,
        sensor_id: int,
        processor_specs: List[ProcessorSpec],
        session_config: a121.SessionConfig,
        noise_calibration: NoiseCalibration,
        close_range_calibration: Optional[CloseRangeCalibration],
    ) -> int:
        _, _, ref_temp = self._recorded_threshold_calibration(
            sensor_id,
            processor_specs,
            session_config,
            noise_calibration,
            close_range_calibration,
        )
        return ref_temp


def _get_group_items(group: h5py.Group) -> list[npt.NDArray[Any]]:
    group_items = []

    i = 0
    while True:
        try:
            v = group[f"index_{i}"][()]
        except KeyError:
            break

        group_items.append(v)
        i += 1
    return group_items


############################
# For migration purpose only
############################


@attrs.mutable(kw_only=True)
class _DetectorContext_v0(AlgoBase):
    single_sensor_contexts: Optional[Dict[int, _SingleSensorContext_v0]] = attrs.field(
        default=None
    )
    _GROUP_NAME = "sensor_id_"

    @property
    def sensor_ids(self) -> Optional[list[int]]:
        if self.single_sensor_contexts:
            return list(self.single_sensor_contexts.keys())
        else:
            return None

    def to_h5(self, group: h5py.Group) -> None:
        if self.single_sensor_contexts is not None:
            for sensor_id, context in self.single_sensor_contexts.items():
                context.to_h5(group.create_group(self._GROUP_NAME + str(sensor_id)))

    @classmethod
    def from_h5(cls, group: h5py.Group) -> _DetectorContext_v0:
        context_dict = {}

        for key in group.keys():  # noqa: SIM118
            if cls._GROUP_NAME in key:
                sensor_id = int(key.split("_")[-1])
                context_dict[sensor_id] = _SingleSensorContext_v0.from_h5(group[key])

        return _DetectorContext_v0(single_sensor_contexts=context_dict)

    @staticmethod
    def _get_calibrate_noise_session_config(
        session_config: a121.SessionConfig, sensor_ids: List[int]
    ) -> a121.SessionConfig:
        noise_session_config = copy.deepcopy(session_config)

        for sensor_id in sensor_ids:
            for group in noise_session_config.groups:
                group[sensor_id].sweeps_per_frame = 1
                # Set num_points to a high number to get sufficient number of data points to
                # estimate the standard deviation. Extra num_points for step_length = 1 together
                # with profile = 5 due to filter margin and cropping
                if any(
                    ss.step_length == 1 and ss.profile == a121.Profile.PROFILE_5
                    for ss in group[sensor_id].subsweeps
                ):
                    num_points = 352
                else:
                    num_points = 220
                for subsweep in group[sensor_id].subsweeps:
                    subsweep.enable_tx = False
                    subsweep.step_length = 1
                    subsweep.start_point = 0
                    subsweep.num_points = num_points

        return noise_session_config

    def migrate(self) -> DetectorContext:
        if not self.single_sensor_contexts:
            return DetectorContext()

        base_step_length_m = 0.00250227400101721
        has_close_range = any(
            [
                ctx.extra_context.close_range_frames is not None
                for ctx in self.single_sensor_contexts.values()
            ]
        )
        has_recorded_threshold = any(
            [
                ctx.extra_context.recorded_threshold_frames is not None
                for ctx in self.single_sensor_contexts.values()
            ]
        )

        # Session config
        session_config_used_during_calibration = list(self.single_sensor_contexts.values())[
            0
        ].session_config_used_during_calibration

        assert session_config_used_during_calibration is not None

        # Metadata
        metadata = a121.Metadata(
            frame_data_length=0,
            sweep_data_length=0,
            subsweep_data_offset=np.array([0]),
            subsweep_data_length=np.array([0]),
            calibration_temperature=25,
            tick_period=0,
            base_step_length_m=base_step_length_m,
            max_sweep_rate=0,
            high_speed_mode=False,
        )
        extended_metadata = [
            {k: metadata for k in group} for group in session_config_used_during_calibration.groups
        ]

        # Offset result
        offset_result = [
            {
                k: a121.Result(
                    data_saturated=False,
                    frame_delayed=False,
                    calibration_needed=False,
                    temperature=ctx.reference_temperature
                    if ctx.reference_temperature is not None
                    else 25,
                    frame=ctx.extra_context.offset_frames[0][0]
                    if ctx.extra_context.offset_frames is not None
                    else np.array([]),
                    tick=0,
                    context=a121._core.entities.ResultContext(
                        metadata=a121.Metadata(
                            frame_data_length=50,
                            sweep_data_length=50,
                            subsweep_data_offset=np.array([0]),
                            subsweep_data_length=np.array([50]),
                            calibration_temperature=ctx.reference_temperature
                            if ctx.reference_temperature is not None
                            else 25,
                            tick_period=0,
                            base_step_length_m=base_step_length_m,
                            max_sweep_rate=168,
                            high_speed_mode=False,
                        ),
                        ticks_per_second=1000,
                    ),
                )
                for k, ctx in self.single_sensor_contexts.items()
            }
        ]

        # Noise result
        sensor_ids = self.sensor_ids
        assert sensor_ids is not None

        noise_session_config = _DetectorContext_v0._get_calibrate_noise_session_config(
            session_config_used_during_calibration, sensor_ids
        )
        noise_result = []

        for group_idx, group in enumerate(noise_session_config.groups):
            group_result = {}
            frame_offset = {}
            for sensor_id, sensor_config in group.items():
                noise_frames = self.single_sensor_contexts[sensor_id].extra_context.noise_frames
                assert noise_frames is not None
                if sensor_id not in frame_offset:
                    frame_offset[sensor_id] = 0
                result_frame: List[npt.NDArray[np.complex128]] = []
                num_points = []
                subsweep_data_offset = []
                total_num_points = 0
                for ss in sensor_config.subsweeps:
                    subsweep_data_offset.append(total_num_points)
                    total_num_points += ss.num_points
                    result_frame.extend(
                        noise_frames[group_idx][0][
                            frame_offset[sensor_id] : ss.num_points
                            * sensor_config.sweeps_per_frame
                        ]
                    )
                    frame_offset[sensor_id] += ss.num_points * sensor_config.sweeps_per_frame
                    num_points.append(ss.num_points)
                result = a121.Result(
                    data_saturated=False,
                    frame_delayed=False,
                    calibration_needed=False,
                    temperature=27,
                    frame=np.array(result_frame),
                    tick=0,
                    context=a121._core.entities.ResultContext(
                        metadata=a121.Metadata(
                            frame_data_length=np.sum(num_points) * sensor_config.sweeps_per_frame,
                            sweep_data_length=np.sum(num_points),
                            subsweep_data_offset=np.array(subsweep_data_offset),
                            subsweep_data_length=np.array(num_points),
                            calibration_temperature=27,
                            tick_period=0,
                            base_step_length_m=base_step_length_m,
                            max_sweep_rate=168,
                            high_speed_mode=False,
                        ),
                        ticks_per_second=1000,
                    ),
                )

                group_result[sensor_id] = result

            noise_result.append(group_result)

        # Close range result
        if has_close_range:
            close_range_result = []

            for group in session_config_used_during_calibration.groups:
                group_result = {}
                frame_offset = {}
                for sensor_id, sensor_config in group.items():
                    close_range_frames = self.single_sensor_contexts[
                        sensor_id
                    ].extra_context.close_range_frames
                    assert close_range_frames is not None
                    if sensor_id not in frame_offset:
                        frame_offset[sensor_id] = 0
                    result_frame = []
                    num_points = []
                    subsweep_data_offset = []
                    total_num_points = 0
                    for ss in sensor_config.subsweeps:
                        subsweep_data_offset.append(total_num_points)
                        total_num_points += ss.num_points
                        result_frame.extend(
                            close_range_frames[0][0][:][
                                frame_offset[sensor_id] : ss.num_points
                                * sensor_config.sweeps_per_frame
                            ]
                        )
                        frame_offset[sensor_id] += ss.num_points * sensor_config.sweeps_per_frame
                        num_points.append(ss.num_points)
                    result = a121.Result(
                        data_saturated=False,
                        frame_delayed=False,
                        calibration_needed=False,
                        temperature=25,
                        frame=np.array(result_frame),
                        tick=0,
                        context=a121._core.entities.ResultContext(
                            metadata=a121.Metadata(
                                frame_data_length=np.sum(num_points)
                                * sensor_config.sweeps_per_frame,
                                sweep_data_length=np.sum(num_points),
                                subsweep_data_offset=np.array(subsweep_data_offset),
                                subsweep_data_length=np.array(num_points),
                                calibration_temperature=25,
                                tick_period=0,
                                base_step_length_m=base_step_length_m,
                                max_sweep_rate=168,
                                high_speed_mode=False,
                            ),
                            ticks_per_second=1000,
                        ),
                    )

                    group_result[sensor_id] = result

                close_range_result.append(group_result)

        # Recorded threshold result
        if has_recorded_threshold:
            recorded_threshold_result = []
            rec_thres_frames = list(self.single_sensor_contexts.values())[
                0
            ].extra_context.recorded_threshold_frames
            assert rec_thres_frames is not None
            num_rec_sweeps = len(rec_thres_frames[0])

            for i in range(num_rec_sweeps):
                group_results = []
                for group_idx, group in enumerate(session_config_used_during_calibration.groups):
                    group_result = {}
                    frame_offset = {}
                    for sensor_id, sensor_config in group.items():
                        recorded_threshold_frames = self.single_sensor_contexts[
                            sensor_id
                        ].extra_context.recorded_threshold_frames
                        assert recorded_threshold_frames is not None
                        if sensor_id not in frame_offset:
                            frame_offset[sensor_id] = 0
                        result_frame = []
                        num_points = []
                        subsweep_data_offset = []
                        total_num_points = 0
                        for ss in sensor_config.subsweeps:
                            subsweep_data_offset.append(total_num_points)
                            total_num_points += ss.num_points
                            result_frame.extend(
                                recorded_threshold_frames[group_idx][i][:][
                                    frame_offset[sensor_id] : ss.num_points
                                    * sensor_config.sweeps_per_frame
                                ]
                            )
                            frame_offset[sensor_id] += (
                                ss.num_points * sensor_config.sweeps_per_frame
                            )
                            num_points.append(ss.num_points)

                        result = a121.Result(
                            data_saturated=False,
                            frame_delayed=False,
                            calibration_needed=False,
                            temperature=25,
                            frame=np.array(result_frame),
                            tick=0,
                            context=a121._core.entities.ResultContext(
                                metadata=a121.Metadata(
                                    frame_data_length=np.sum(num_points)
                                    * sensor_config.sweeps_per_frame,
                                    sweep_data_length=np.sum(num_points),
                                    subsweep_data_offset=np.array(subsweep_data_offset),
                                    subsweep_data_length=np.array(num_points),
                                    calibration_temperature=25,
                                    tick_period=0,
                                    base_step_length_m=base_step_length_m,
                                    max_sweep_rate=168,
                                    high_speed_mode=False,
                                ),
                                ticks_per_second=1000,
                            ),
                        )

                        group_result[sensor_id] = result
                    group_results.append(group_result)
                recorded_threshold_result.append(group_results)

        # Sensor calibrations
        sensor_calibrations: Dict[int, Optional[a121.SensorCalibration]] = {
            k: v.sensor_calibration for k, v in self.single_sensor_contexts.items()
        }
        sensor_calibrations_re: Dict[int, a121.SensorCalibration] = (
            {k: v for k, v in sensor_calibrations.items() if v is not None}
            if all([x is not None for x in sensor_calibrations.values()])
            else {}
        )
        return DetectorContext(
            offset_calibration=OffsetCalibration(results=offset_result),
            noise_calibration=NoiseCalibration(results=noise_result),
            close_range_calibration=CloseRangeCalibration(
                extended_metadata=extended_metadata,
                results=close_range_result,
                sensor_calibrations=sensor_calibrations_re,
            )
            if has_close_range
            else None,
            recorded_threshold_calibration=RecordedThresholdCalibration(
                extended_metadata=extended_metadata, results=recorded_threshold_result
            )
            if has_recorded_threshold
            else None,
            sensor_ids=list(self.single_sensor_contexts.keys()),
            session_config_used_during_calibration=session_config_used_during_calibration,
        )


@attrs.mutable(kw_only=True)
class _SingleSensorExtraContext_v0(AlgoBase):
    offset_frames: Optional[List[List[npt.NDArray[np.complex128]]]] = attrs.field(
        default=None, eq=attrs_optional_ndarray_isclose
    )
    noise_frames: Optional[List[List[npt.NDArray[np.complex128]]]] = attrs.field(
        default=None, eq=attrs_optional_ndarray_isclose
    )
    close_range_frames: Optional[List[List[npt.NDArray[np.complex128]]]] = attrs.field(
        default=None, eq=attrs_optional_ndarray_isclose
    )
    recorded_threshold_frames: Optional[List[List[npt.NDArray[np.complex128]]]] = attrs.field(
        default=None, eq=attrs_optional_ndarray_isclose
    )


@attrs.mutable(kw_only=True)
class _SingleSensorContext_v0(AlgoBase):
    loopback_peak_location_m: Optional[float] = attrs.field(default=None)
    direct_leakage: Optional[npt.NDArray[np.complex128]] = attrs.field(
        default=None, eq=attrs_optional_ndarray_isclose
    )
    phase_jitter_comp_reference: Optional[npt.NDArray[np.float64]] = attrs.field(
        default=None, eq=attrs_optional_ndarray_isclose
    )
    recorded_thresholds_mean_sweep: Optional[List[npt.NDArray[np.float64]]] = attrs.field(
        default=None, eq=attrs_optional_ndarray_isclose
    )
    recorded_thresholds_noise_std: Optional[List[List[np.float64]]] = attrs.field(
        default=None, eq=attrs_optional_ndarray_isclose
    )
    bg_noise_std: Optional[List[List[float]]] = attrs.field(
        default=None, eq=attrs_optional_ndarray_isclose
    )
    session_config_used_during_calibration: Optional[a121.SessionConfig] = attrs.field(
        default=None
    )
    reference_temperature: Optional[int] = attrs.field(default=None)
    sensor_calibration: Optional[a121.SensorCalibration] = attrs.field(default=None)
    extra_context: _SingleSensorExtraContext_v0 = attrs.field(factory=_SingleSensorExtraContext_v0)
    # TODO: Make recorded_thresholds Optional[List[Optional[npt.NDArray[np.float64]]]]

    def to_h5(self, group: h5py.Group) -> None:
        for k, v in attrs.asdict(self, recurse=False).items():
            if k in [
                "recorded_thresholds_mean_sweep",
                "recorded_thresholds_noise_std",
                "bg_noise_std",
                "extra_context",
            ]:
                continue

            if v is None:
                continue

            if isinstance(v, a121.SessionConfig):
                _create_h5_string_dataset(group, k, v.to_json())
            elif isinstance(v, a121.SensorCalibration):
                sensor_calibration_group = group.create_group("sensor_calibration")
                v.to_h5(sensor_calibration_group)
            elif isinstance(v, (np.ndarray, float, int, np.integer)):
                group.create_dataset(k, data=v, track_times=False)
            else:
                msg = f"Unexpected {type(self).__name__} field '{k}' of type '{type(v)}'"
                raise RuntimeError(msg)

        if self.recorded_thresholds_mean_sweep is not None:
            recorded_thresholds_mean_sweep_group = group.create_group(
                "recorded_thresholds_mean_sweep"
            )

            for i, v in enumerate(self.recorded_thresholds_mean_sweep):
                recorded_thresholds_mean_sweep_group.create_dataset(
                    f"index_{i}", data=v, track_times=False
                )

        if self.recorded_thresholds_noise_std is not None:
            recorded_thresholds_std_group = group.create_group("recorded_thresholds_noise_std")

            for i, v in enumerate(self.recorded_thresholds_noise_std):
                recorded_thresholds_std_group.create_dataset(
                    f"index_{i}", data=v, track_times=False
                )

        if self.bg_noise_std is not None:
            bg_noise_std_group = group.create_group("bg_noise_std")

            for i, v in enumerate(self.bg_noise_std):
                bg_noise_std_group.create_dataset(f"index_{i}", data=v, track_times=False)

        extra_group = group.create_group("extra_context")

        if self.extra_context.offset_frames is not None:
            offset_frames_group = extra_group.create_group("offset_frames")

            for i, v in enumerate(self.extra_context.offset_frames):
                offset_frames_group.create_dataset(f"index_{i}", data=v, track_times=False)

        if self.extra_context.noise_frames is not None:
            noise_frames_group = extra_group.create_group("noise_frames")

            for i, v in enumerate(self.extra_context.noise_frames):
                noise_frames_group.create_dataset(f"index_{i}", data=v, track_times=False)

        if self.extra_context.close_range_frames is not None:
            close_range_frames_group = extra_group.create_group("close_range_frames")

            for i, v in enumerate(self.extra_context.close_range_frames):
                close_range_frames_group.create_dataset(f"index_{i}", data=v, track_times=False)

        if self.extra_context.recorded_threshold_frames is not None:
            recorded_threshold_frames_group = extra_group.create_group("recorded_threshold_frames")

            for i, v in enumerate(self.extra_context.recorded_threshold_frames):
                recorded_threshold_frames_group.create_dataset(
                    f"index_{i}", data=v, track_times=False
                )

    @classmethod
    def from_h5(cls, group: h5py.Group) -> _SingleSensorContext_v0:
        context_dict: Dict[str, Any] = {}
        context_dict["extra_context"] = {}

        unknown_keys = set(group.keys()) - set(attrs.fields_dict(_SingleSensorContext_v0).keys())
        if unknown_keys:
            msg = f"Unknown field(s) in stored context: {unknown_keys}"
            raise Exception(msg)

        field_map = {
            "loopback_peak_location_m": None,
            "direct_leakage": None,
            "reference_temperature": None,
            "phase_jitter_comp_reference": None,
            "session_config_used_during_calibration": a121.SessionConfig.from_json,
        }
        for k, func in field_map.items():
            try:
                v = group[k][()]
            except KeyError:
                continue

            context_dict[k] = func(v) if func else v

        if "recorded_thresholds_mean_sweep" in group:
            mean_sweeps = _get_group_items(group["recorded_thresholds_mean_sweep"])
            context_dict["recorded_thresholds_mean_sweep"] = mean_sweeps

        if "recorded_thresholds_noise_std" in group:
            noise_stds = _get_group_items(group["recorded_thresholds_noise_std"])
            context_dict["recorded_thresholds_noise_std"] = noise_stds

        if "bg_noise_std" in group:
            bg_noise_std = _get_group_items(group["bg_noise_std"])
            context_dict["bg_noise_std"] = bg_noise_std

        if "sensor_calibration" in group:
            context_dict["sensor_calibration"] = a121.SensorCalibration.from_h5(
                group["sensor_calibration"]
            )

        if "extra_context" in group:
            extra_group = group["extra_context"]

            if "offset_frames" in extra_group:
                offset_frames = _get_group_items(extra_group["offset_frames"])
                context_dict["extra_context"]["offset_frames"] = offset_frames

            if "noise_frames" in extra_group:
                noise_frames = _get_group_items(extra_group["noise_frames"])
                context_dict["extra_context"]["noise_frames"] = noise_frames

            if "close_range_frames" in extra_group:
                close_range_frames = _get_group_items(extra_group["close_range_frames"])
                context_dict["extra_context"]["close_range_frames"] = close_range_frames

            if "recorded_threshold_frames" in extra_group:
                recorded_threshold_frames = _get_group_items(
                    extra_group["recorded_threshold_frames"]
                )
                context_dict["extra_context"]["recorded_threshold_frames"] = (
                    recorded_threshold_frames
                )

        context_dict["extra_context"] = _SingleSensorExtraContext_v0(
            **context_dict["extra_context"]
        )

        return _SingleSensorContext_v0(**context_dict)


detector_context_timeline = (
    tm.start(_DetectorContext_v0)
    .load(str, _DetectorContext_v0.from_json, fail=[])
    .load(h5py.Group, _DetectorContext_v0.from_h5, fail=[])
    .nop()
    .epoch(DetectorContext, _DetectorContext_v0.migrate, fail=[])
    .load(str, DetectorContext.from_json, fail=[])
    .load(
        h5py.Group,
        lambda x: opser.deserialize(x, DetectorContext),
        fail=[opser.core.LoadErrorGroup],
    )
)
